@@ -46,7 +46,10 @@ GetOptions(\%args, qw(
     static has-libtommath has-libatomic_ops
     has-dyncall has-libffi pkgconfig=s
     build=s host=s big-endian jit! enable-jit
-    prefix=s bindir=s libdir=s mastdir=s make-install asan ubsan valgrind telemeh show-autovect show-autovect-failed:s),
+    prefix=s bindir=s libdir=s mastdir=s
+    relocatable make-install asan ubsan
+    valgrind telemeh show-autovect
+    show-autovect-failed:s),
 
     'no-optimize|nooptimize' => sub { $args{optimize} = 0 },
     'no-debug|nodebug' => sub { $args{debug} = 0 },
@@ -68,7 +71,7 @@ Please specify another installation target by using --prefix=PATH.
 ENOTTOCWD
 }
 
-# Override default target directories with command line argumets
+# Override default target directories with command line arguments
 my @target_dirs = qw{bindir libdir mastdir};
 for my $target (@target_dirs) {
     $config{$target} = $args{$target} if $args{$target};
@@ -91,6 +94,11 @@ if (-d '.git') {
 $args{optimize}     = 3 if not defined $args{optimize} or $args{optimize} eq "";
 $args{debug}        = 3 if defined $args{debug} and $args{debug} eq "";
 
+# Relocatability is not supported on AIX and OpenBSD.
+if ( $args{relocatable} && ($^O eq 'aix' || $^O eq 'openbsd') ) {
+    hardfail('Relocatability is not supported on ' . $^O .
+    ".\n    Leave off the --relocatable flag to do a non-relocatable build.");
+}
 
 for (qw(coverage instrument static big-endian has-libtommath has-sha has-libuv
         has-libatomic_ops asan ubsan valgrind show-vec)) {
@@ -167,6 +175,43 @@ $config{ldinstflags}  = $config{ccinstflags} unless defined $config{ldinstflags}
 
 $config{as}           = $config{cc} unless defined $config{as};
 
+# If we're in macOS, let's verify that the toolchain is consistent.
+if ($^O eq 'darwin') {
+    my $gnu_count = 0;
+    my $gnu_toolchain = (exists $args{toolchain}) && ($args{toolchain} eq 'gnu');
+
+    unless ($gnu_toolchain) {
+        # When XCode toolchain is used then force use of XCode's make if
+        # available.
+        $config{make} = '/usr/bin/make' if -x '/usr/bin/make'; 
+    }
+
+    # Here are the tools that seem to cause trouble.
+    # If you see other ones, please add them to this list.
+    my @check_tools = qw/ar cc ld/;
+    for my $tool (map { `which $_` } @config{@check_tools}) {
+        chomp $tool;
+        system "grep -b 'gnu' '$tool'"; # Apple utilities don't match `gnu`
+        if ($? == 0) {
+            $gnu_count += 1;
+        }
+    }
+
+    ## For a GNU toolchain, make sure that they're all GNU.
+    if ($gnu_toolchain && $gnu_count != scalar @check_tools) {
+        print "\nNot all tools in the toolchain are GNU. Please correct this and retry.\n"
+            . "See README.markdown for more details.\n\n";
+        exit -1;
+    }
+
+    ## Otherwise, make sure that none of them are GNU
+    elsif (!$gnu_toolchain && $gnu_count != 0) {
+        print "\nGNU tools detected, despite this not being a GNU-oriented build.\n"
+            ." Please correct this and retry. See README.markdown for more details.\n\n";
+        exit -1;
+    }
+}
+
 # Probe the compiler.
 build::probe::compiler_usability(\%config, \%defaults);
 
@@ -178,6 +223,11 @@ if ($config{cc} eq 'gcc' && !$config{can_specific_werror}) {
     $config{ccmiscflags} =~ s/^ +$//;
 }
 
+# Disable RETGUARD on OpenBSD, since it breaks the legojit.
+if ($^O eq 'openbsd' && $args{jit} && $config{cc} eq 'clang') {
+	$config{ccmiscflags} .= ' -fno-ret-protector';
+}
+
 # Set the remaining ldmiscflags. Do this after probing for gcc -Werror probe to not miss that change for the linker.
 $config{ldmiscflags}  = $config{ccmiscflags} unless defined $config{ldmiscflags};
 
@@ -212,20 +262,12 @@ if (-e "$config{pkgconfig}") {
 
 # conditionally set include dirs and install rules
 $config{cincludes} = '' unless defined $config{cincludes};
+$config{lincludes} = '' unless defined $config{lincludes};
 $config{install}   = '' unless defined $config{install};
 if ($args{'has-libuv'}) {
     $defaults{-thirdparty}->{uv} = undef;
     unshift @{$config{usrlibs}}, 'uv';
-    if ($config{pkgconfig_works}) {
-        my $result = `$config{pkgconfig} --cflags libuv`;
-        if ( $? == 0 ) {
-            $result =~ s/\n/ /g;
-            $config{cincludes} .= ' ' . "$result";
-            print("Adding extra include for libuv: $result\n");
-        } else {
-            print("Error occured when running $config{pkgconfig} --cflags libuv.\n");
-        }
-    }
+    setup_native_library('libuv') if $config{pkgconfig_works};
 }
 else {
     $config{cincludes} .= ' ' . $defaults{ccinc} . '3rdparty/libuv/include'
@@ -239,16 +281,7 @@ else {
 if ($args{'has-libatomic_ops'}) {
     $defaults{-thirdparty}->{lao} = undef;
     unshift @{$config{usrlibs}}, 'atomic_ops';
-    if ($config{pkgconfig_works}) {
-        my $result = `$config{pkgconfig} --cflags atomic_ops`;
-        if ( $? == 0 ) {
-            $result =~ s/\n/ /g;
-            $config{cincludes} .= ' ' . "$result";
-            print("Adding extra include for atomic_ops: $result\n");
-        } else {
-            print("Error occured when running $config{pkgconfig} --cflags atomic_ops.\n");
-        }
-    }
+    setup_native_library('atomic_ops') if $config{pkgconfig_works};
 }
 else {
     $config{cincludes} .= ' ' . $defaults{ccinc} . '3rdparty/libatomicops/src';
@@ -277,6 +310,14 @@ else {
 if ($args{'has-libtommath'}) {
     $defaults{-thirdparty}->{tom} = undef;
     unshift @{$config{usrlibs}}, 'tommath';
+    if (not $config{crossconf}) {
+        if (index($config{cincludes}, '-I/usr/local/include') == -1) {
+            $config{cincludes} = join(' ', $config{cincludes}, '-I/usr/local/include');
+        }
+        if (index($config{lincludes}, '-L/usr/local/lib') == -1) {
+            $config{lincludes} = join(' ', $config{lincludes}, '-L/usr/local/lib');
+        }
+    }
 }
 else {
     $config{cincludes} .= ' ' . $defaults{ccinc} . '3rdparty/libtommath';
@@ -292,24 +333,7 @@ if ($args{'has-libffi'}) {
     $defaults{-thirdparty}->{dcb} = undef;
     $defaults{-thirdparty}->{dl}  = undef;
     if ($config{pkgconfig_works}) {
-        my $result_cflags = `$config{pkgconfig} --cflags libffi`;
-        if ( $? == 0 ) {
-            $result_cflags =~ s/\n/ /g;
-            $config{cincludes} .= " $result_cflags";
-            print("Adding extra include for libffi: $result_cflags\n");
-        }
-        else {
-            print("Error occured when running $config{pkgconfig} --cflags libffi.\n");
-        }
-        my $result_libs = `$config{pkgconfig} --libs libffi`;
-        if ( $? == 0 ) {
-            $result_libs =~ s/\n/ /g;
-            $config{ldusr} .= " $result_libs";
-            print("Adding extra libs for libffi: $result_libs\n");
-        }
-        else {
-            print("Error occured when running $config{pkgconfig} --libs libffi.\n");
-        }
+        setup_native_library('libffi');
     }
     elsif ($^O eq 'solaris') {
         my ($first) = map { m,(.+)/ffi\.h$, && "/$1"  } grep { m,/ffi\.h$, } `pkg contents libffi`;
@@ -328,6 +352,14 @@ elsif ($args{'has-dyncall'}) {
     $defaults{-thirdparty}->{dcb} = undef;
     $defaults{-thirdparty}->{dl}  = undef;
     $config{nativecall_backend} = 'dyncall';
+    if (not $config{crossconf}) {
+        if (index($config{cincludes}, '-I/usr/local/include') == -1) {
+            $config{cincludes} = join(' ', $config{cincludes}, '-I/usr/local/include');
+        }
+        if (index($config{lincludes}, '-L/usr/local/lib') == -1) {
+            $config{lincludes} = join(' ', $config{lincludes}, '-L/usr/local/lib');
+        }
+    }
 }
 else {
     $config{nativecall_backend} = 'dyncall';
@@ -342,6 +374,7 @@ else {
 
 # mangle library names
 $config{ldlibs} = join ' ',
+    $config{lincludes},
     (map { sprintf $config{ldusr}, $_; } @{$config{usrlibs}}),
     (map { sprintf $config{ldsys}, $_; } @{$config{syslibs}});
 $config{ldlibs} = ' -lasan ' . $config{ldlibs} if $args{asan} && $^O ne 'darwin' && $config{cc} ne 'clang';
@@ -398,13 +431,23 @@ $config{cflags} = join ' ', uniq(@cflags);
 my @ldflags = ($config{ldmiscflags});
 push @ldflags, $config{ldoptiflags}  if $args{optimize};
 push @ldflags, $config{lddebugflags} if $args{debug};
-push @ldflags, $config{ldinstflags}       if $args{instrument};
+push @ldflags, $config{ldinstflags}  if $args{instrument};
 push @ldflags, $config{ld_covflags}  if $args{coverage};
-push @ldflags, $config{ldrpath}           if not $args{static} and $config{prefix} ne '/usr';
-push @ldflags, '-fsanitize=address' if $args{asan};
-push @ldflags, $ENV{LDFLAGS}  if $ENV{LDFLAGS};
+if (not $args{static} and $config{prefix} ne '/usr') {
+    push @ldflags, $config{ldrpath_relocatable} if  $args{relocatable};
+    push @ldflags, $config{ldrpath}             if !$args{relocatable};
+}
+push @ldflags, '-fsanitize=address'  if $args{asan};
+push @ldflags, $ENV{LDFLAGS}         if $ENV{LDFLAGS};
 $config{ldflags} = join ' ', @ldflags;
 
+$config{moarshared} = '';
+# Switch shared lib compiler flags in relocatable case.
+if (not $args{static} and $config{prefix} ne '/usr') {
+    $config{moarshared} = $config{moarshared_relocatable}   if  $args{relocatable};
+    $config{moarshared} = $config{moarshared_norelocatable} if !$args{relocatable};
+}
+
 # setup library names
 $config{moarlib} = sprintf $config{lib}, $NAME;
 $config{moardll} = sprintf $config{dll}, $NAME;
@@ -484,6 +527,12 @@ if ($config{cc} eq 'cl') {
 build::probe::C_type_bool(\%config, \%defaults);
 build::probe::computed_goto(\%config, \%defaults);
 build::probe::pthread_yield(\%config, \%defaults);
+build::probe::check_fn_malloc_trim(\%config, \%defaults);
+if ($^O eq 'aix') {
+    build::probe::numbits(\%config, \%defaults);
+    $config{ldflags} = join(',', $config{ldflags}, '-bmaxdata:0x80000000')
+        if $config{arch_bits} == 32;
+}
 build::probe::rdtscp(\%config, \%defaults);
 
 my $order = $config{be} ? 'big endian' : 'little endian';
@@ -781,8 +830,8 @@ sub generate {
             # In-between slashes in makefiles need to be backslashes on Windows.
             # Double backslashes in config.c, beause these are in qq-strings.
             my $bs = $dest =~ /Makefile/ ? '\\' : '\\\\';
-			$line =~ s/(\w|\.|\w\:|\$\(PREFIX\))\/(?=\w|\.|\*)/$1$bs/g;
-			$line =~ s/(\w|\.|\w\:|\$\(PREFIX\))\\(?=\w|\.|\*)/$1$bs/g if $bs eq '\\\\';
+            $line =~ s/(\w|\.|\w\:|\$\(PREFIX\))\/(?=\w|\.|\*)/$1$bs/g;
+            $line =~ s/(\w|\.|\w\:|\$\(PREFIX\))\\(?=\w|\.|\*)/$1$bs/g if $bs eq '\\\\';
 
             # gmake doesn't like \*
             $line =~ s/(\w|\.|\w\:|\$\(PREFIX\))\\\*/$1\\\\\*/g
@@ -851,6 +900,33 @@ sub wsl_bash_on_win {
     return ((readline $fh) =~ /\A\d\.\d\.\d-\d+-Microsoft\s*\z/) ? 1 : 0;
 }
 
+# set up cincludes and lincludes flags for a native library
+sub setup_native_library {
+    my $library = shift;
+    my $result_cflags = `$config{pkgconfig} --cflags $library`;
+    if ( $? == 0 ) {
+        $result_cflags =~ s/\n//g;
+        if (index($config{cincludes}, $result_cflags) == -1) {
+            $config{cincludes} = join(' ', $config{cincludes}, $result_cflags);
+            print("Adding extra include for $library: $result_cflags\n");
+        }
+    }
+    else {
+        print("Error occured when running $config{pkgconfig} --cflags $library.\n");
+    }
+    my $result_libs = `$config{pkgconfig} --libs-only-L $library`;
+    if ( $? == 0 ) {
+        $result_libs =~ s/\n//g;
+        if (index($config{lincludes}, $result_libs) == -1) {
+            $config{lincludes} = join(' ', $config{lincludes}, $result_libs);
+            print("Adding extra libs for $library: $result_libs\n");
+        }
+    }
+    else {
+        print("Error occured when running $config{pkgconfig} --libs-only-L $library.\n");
+    }
+}
+
 __END__
 
 =head1 SYNOPSIS
@@ -861,7 +937,7 @@ __END__
                    [--toolchain <toolchain>] [--compiler <compiler>]
                    [--ar <ar>] [--cc <cc>] [--ld <ld>] [--make <make>]
                    [--debug] [--optimize] [--instrument]
-                   [--static] [--prefix]
+                   [--static] [--prefix <path>] [--relocatable]
                    [--has-libtommath] [--has-sha] [--has-libuv]
                    [--has-libatomic_ops]
                    [--asan] [--ubsan] [--no-jit]
@@ -870,8 +946,8 @@ __END__
     ./Configure.pl --build <build-triple> --host <host-triple>
                    [--ar <ar>] [--cc <cc>] [--ld <ld>] [--make <make>]
                    [--debug] [--optimize] [--instrument]
-                   [--static] [--big-endian] [--prefix]
-                   [--make-install]
+                   [--static] [--prefix <path>] [--relocatable]
+                   [--big-endian] [--make-install]
 
 =head2 Use of environment variables
 
@@ -1003,6 +1079,12 @@ builds, the byte order is auto-detected.
 Install files in subdirectory /bin, /lib and /include of the supplied path.
 The default prefix is "install" if this option is not passed.
 
+=item --relocatable
+
+Search for the libmoar library relative to the executable path.
+(On AIX and OpenBSD MoarVM can not be built relocatable, since both OS'
+miss the necessary mechanism to make this work.)
+
 =item --bindir
 
 Install executable files in the supplied path.  The default is
