@@ -3,9 +3,9 @@
 /* Gets the HOW (meta-object), which may be lazily deserialized. */
 MVMObject * MVM_6model_get_how(MVMThreadContext *tc, MVMSTable *st) {
     MVMObject *HOW = st->HOW;
-    if (!HOW)
+    if (!HOW && st->HOW_sc)
         st->HOW = HOW = MVM_sc_get_object(tc, st->HOW_sc, st->HOW_idx);
-    return HOW;
+    return HOW ? HOW : tc->instance->VMNull;
 }
 
 /* Gets the HOW (meta-object), which may be lazily deserialized, through the
@@ -18,7 +18,7 @@ MVMObject * MVM_6model_get_how_obj(MVMThreadContext *tc, MVMObject *obj) {
 static MVMObject * get_method_cache(MVMThreadContext *tc, MVMSTable *st) {
     if (!st->method_cache)
         MVM_serialization_finish_deserialize_method_cache(tc, st);
-   return st->method_cache;
+    return st->method_cache;
 }
 
 /* Locates a method by name, checking in the method cache only. */
@@ -40,6 +40,7 @@ typedef struct {
     MVMObject   *obj;
     MVMString   *name;
     MVMRegister *res;
+    MVMint64     throw_if_not_found;
 } FindMethodSRData;
 static void die_over_missing_method(MVMThreadContext *tc, MVMObject *obj, MVMString *name) {
     MVMObject *handler = MVM_hll_current(tc)->method_not_found_error;
@@ -57,44 +58,55 @@ static void die_over_missing_method(MVMThreadContext *tc, MVMObject *obj, MVMStr
         char *waste[] = { c_name, NULL };
         MVM_exception_throw_adhoc_free(tc, waste,
             "Cannot find method '%s' on object of type %s",
-            c_name, STABLE(obj)->debug_name);
+            c_name, MVM_6model_get_debug_name(tc, obj));
     }
 }
 static void late_bound_find_method_return(MVMThreadContext *tc, void *sr_data) {
     FindMethodSRData *fm = (FindMethodSRData *)sr_data;
     if (MVM_is_null(tc, fm->res->o) || !IS_CONCRETE(fm->res->o)) {
-        MVMObject *obj  = fm->obj;
-        MVMString *name = fm->name;
-        MVM_free(fm);
-        die_over_missing_method(tc, obj, name);
+        if (fm->throw_if_not_found) {
+            MVMObject *obj  = fm->obj;
+            MVMString *name = fm->name;
+            MVM_free(fm);
+            die_over_missing_method(tc, obj, name);
+        }
+        else {
+            fm->res->o = tc->instance->VMNull;
+            MVM_free(fm);
+        }
     }
     else {
         MVM_free(fm);
     }
 }
 static void mark_find_method_sr_data(MVMThreadContext *tc, MVMFrame *frame, MVMGCWorklist *worklist) {
-    FindMethodSRData *fm = (FindMethodSRData *)frame->special_return_data;
+    FindMethodSRData *fm = (FindMethodSRData *)frame->extra->special_return_data;
     MVM_gc_worklist_add(tc, worklist, &fm->obj);
     MVM_gc_worklist_add(tc, worklist, &fm->name);
 }
-void MVM_6model_find_method(MVMThreadContext *tc, MVMObject *obj, MVMString *name, MVMRegister *res) {
-    MVMObject *cache, *HOW, *find_method, *code;
-    MVMCallsite *findmeth_callsite;
+void MVM_6model_find_method(MVMThreadContext *tc, MVMObject *obj, MVMString *name,
+                            MVMRegister *res, MVMint64 throw_if_not_found) {
+    MVMObject *cache = NULL, *HOW = NULL, *find_method = NULL, *code = NULL;
+    MVMCallsite *findmeth_callsite = NULL;
 
     if (MVM_is_null(tc, obj)) {
-        char *c_name  = MVM_string_utf8_encode_C_string(tc, name);
-        char *waste[] = { c_name, NULL };
-        MVM_exception_throw_adhoc_free(tc, waste,
-            "Cannot call method '%s' on a null object",
-             c_name);
+        if (throw_if_not_found) {
+            char *c_name  = MVM_string_utf8_encode_C_string(tc, name);
+            char *waste[] = { c_name, NULL };
+            MVM_exception_throw_adhoc_free(tc, waste,
+                "Cannot call method '%s' on a null object",
+                 c_name);
+        }
+        else {
+            res->o = tc->instance->VMNull;
+            return;
+        }
     }
 
     /* First try to find it in the cache. If we find it, we have a result.
      * If we don't find it, but the cache is authoritative, then error. */
-    MVMROOT(tc, obj, {
-        MVMROOT(tc, name, {
-            cache = get_method_cache(tc, STABLE(obj));
-        });
+    MVMROOT2(tc, obj, name, {
+        cache = get_method_cache(tc, STABLE(obj));
     });
 
     if (cache && IS_CONCRETE(cache)) {
@@ -104,29 +116,34 @@ void MVM_6model_find_method(MVMThreadContext *tc, MVMObject *obj, MVMString *nam
             return;
         }
         if (STABLE(obj)->mode_flags & MVM_METHOD_CACHE_AUTHORITATIVE) {
-            die_over_missing_method(tc, obj, name);
+            if (throw_if_not_found)
+                die_over_missing_method(tc, obj, name);
+            else
+                res->o = tc->instance->VMNull;
             return;
         }
     }
 
     /* Otherwise, need to call the find_method method. We make the assumption
      * that the invocant's meta-object's type is composed. */
-    MVMROOT(tc, obj, {
-        MVMROOT(tc, name, {
-            HOW = MVM_6model_get_how(tc, STABLE(obj));
-            MVMROOT(tc, HOW, {
-                 find_method = MVM_6model_find_method_cache_only(tc, HOW,
-                      tc->instance->str_consts.find_method);
-            });
-        });
+    MVMROOT3(tc, obj, name, HOW, {
+       HOW = MVM_6model_get_how(tc, STABLE(obj));
+       find_method = MVM_6model_find_method_cache_only(tc, HOW,
+            tc->instance->str_consts.find_method);
     });
 
     if (MVM_is_null(tc, find_method)) {
-        char *c_name  = MVM_string_utf8_encode_C_string(tc, name);
-        char *waste[] = { c_name, NULL };
-        MVM_exception_throw_adhoc_free(tc, waste,
-            "Cannot find method '%s': no method cache and no .^find_method",
-             c_name);
+        if (throw_if_not_found) {
+            char *c_name  = MVM_string_utf8_encode_C_string(tc, name);
+            char *waste[] = { c_name, NULL };
+            MVM_exception_throw_adhoc_free(tc, waste,
+                "Cannot find method '%s' on '%s': no method cache and no .^find_method",
+                 c_name, MVM_6model_get_debug_name(tc, obj));
+        }
+        else {
+            res->o = tc->instance->VMNull;
+            return;
+        }
     }
 
     /* Set up the call, using the result register as the target. */
@@ -138,9 +155,9 @@ void MVM_6model_find_method(MVMThreadContext *tc, MVMObject *obj, MVMString *nam
         fm->obj  = obj;
         fm->name = name;
         fm->res  = res;
-        tc->cur_frame->special_return           = late_bound_find_method_return;
-        tc->cur_frame->special_return_data      = fm;
-        tc->cur_frame->mark_special_return_data = mark_find_method_sr_data;
+        fm->throw_if_not_found = throw_if_not_found;
+        MVM_frame_special_return(tc, tc->cur_frame, late_bound_find_method_return,
+            NULL, fm, mark_find_method_sr_data);
     }
     tc->cur_frame->args[0].o = HOW;
     tc->cur_frame->args[1].o = obj;
@@ -154,10 +171,8 @@ MVMint32 MVM_6model_find_method_spesh(MVMThreadContext *tc, MVMObject *obj, MVMS
 
     /* Missed mono-morph; try cache-only lookup. */
 
-    MVMROOT(tc, obj, {
-        MVMROOT(tc, name, {
-            meth = MVM_6model_find_method_cache_only(tc, obj, name);
-        });
+    MVMROOT2(tc, obj, name, {
+        meth = MVM_6model_find_method_cache_only(tc, obj, name);
     });
 
     if (!MVM_is_null(tc, meth)) {
@@ -166,11 +181,12 @@ MVMint32 MVM_6model_find_method_spesh(MVMThreadContext *tc, MVMObject *obj, MVMS
         MVMStaticFrame *sf = tc->cur_frame->static_info;
         uv_mutex_lock(&tc->instance->mutex_spesh_install);
         if (!tc->cur_frame->effective_spesh_slots[ss_idx + 1]) {
-            MVM_ASSIGN_REF(tc, &(sf->common.header),
+            MVMStaticFrameSpesh *spesh = sf->body.spesh;
+            MVM_ASSIGN_REF(tc, &(spesh->common.header),
                            tc->cur_frame->effective_spesh_slots[ss_idx + 1],
                            (MVMCollectable *)meth);
             MVM_barrier();
-            MVM_ASSIGN_REF(tc, &(sf->common.header),
+            MVM_ASSIGN_REF(tc, &(spesh->common.header),
                            tc->cur_frame->effective_spesh_slots[ss_idx],
                            (MVMCollectable *)STABLE(obj));
         }
@@ -180,7 +196,7 @@ MVMint32 MVM_6model_find_method_spesh(MVMThreadContext *tc, MVMObject *obj, MVMS
     }
     else {
         /* Fully late-bound. */
-        MVM_6model_find_method(tc, obj, name, res);
+        MVM_6model_find_method(tc, obj, name, res, 1);
         return 1;
     }
 }
@@ -206,10 +222,8 @@ MVMint64 MVM_6model_can_method_cache_only(MVMThreadContext *tc, MVMObject *obj,
 
     /* Consider the method cache. */
 
-    MVMROOT(tc, obj, {
-        MVMROOT(tc, name, {
-            cache = get_method_cache(tc, STABLE(obj));
-        });
+    MVMROOT2(tc, obj, name, {
+        cache = get_method_cache(tc, STABLE(obj));
     });
 
     if (cache && IS_CONCRETE(cache)) {
@@ -225,15 +239,13 @@ MVMint64 MVM_6model_can_method_cache_only(MVMThreadContext *tc, MVMObject *obj,
 }
 
 void MVM_6model_can_method(MVMThreadContext *tc, MVMObject *obj, MVMString *name, MVMRegister *res) {
-    MVMObject *HOW, *find_method, *code;
-    MVMCallsite *findmeth_callsite;
+    MVMObject *HOW = NULL, *find_method = NULL, *code = NULL;
+    MVMCallsite *findmeth_callsite = NULL;
 
     MVMint64 can_cached;
 
-    MVMROOT(tc, obj, {
-        MVMROOT(tc, name, {
-            can_cached = MVM_6model_can_method_cache_only(tc, obj, name);
-        });
+    MVMROOT2(tc, obj, name, {
+        can_cached = MVM_6model_can_method_cache_only(tc, obj, name);
     });
 
     if (can_cached == 0 || can_cached == 1) {
@@ -243,14 +255,10 @@ void MVM_6model_can_method(MVMThreadContext *tc, MVMObject *obj, MVMString *name
 
     /* If no method in cache and the cache is not authoritative, need to make
      * a late-bound call to find_method. */
-    MVMROOT(tc, obj, {
-        MVMROOT(tc, name, {
-            HOW = MVM_6model_get_how(tc, STABLE(obj));
-            MVMROOT(tc, HOW, {
-                find_method = MVM_6model_find_method_cache_only(tc, HOW,
-                     tc->instance->str_consts.find_method);
-            });
-       });
+    MVMROOT3(tc, obj, name, HOW, {
+        HOW = MVM_6model_get_how(tc, STABLE(obj));
+        find_method = MVM_6model_find_method_cache_only(tc, HOW,
+            tc->instance->str_consts.find_method);
     });
 
     if (MVM_is_null(tc, find_method)) {
@@ -264,8 +272,7 @@ void MVM_6model_can_method(MVMThreadContext *tc, MVMObject *obj, MVMString *name
     code = MVM_frame_find_invokee(tc, find_method, NULL);
     findmeth_callsite = MVM_callsite_get_common(tc, MVM_CALLSITE_ID_FIND_METHOD);
     MVM_args_setup_thunk(tc, res, MVM_RETURN_OBJ, findmeth_callsite);
-    tc->cur_frame->special_return      = late_bound_can_return;
-    tc->cur_frame->special_return_data = res;
+    MVM_frame_special_return(tc, tc->cur_frame, late_bound_can_return, NULL, res, NULL);
     tc->cur_frame->args[0].o = HOW;
     tc->cur_frame->args[1].o = obj;
     tc->cur_frame->args[2].s = name;
@@ -275,16 +282,12 @@ void MVM_6model_can_method(MVMThreadContext *tc, MVMObject *obj, MVMString *name
 /* Checks if an object has a given type, delegating to the type_check or
  * accepts_type methods as needed. */
 static void do_accepts_type_check(MVMThreadContext *tc, MVMObject *obj, MVMObject *type, MVMRegister *res) {
-    MVMObject *HOW, *meth;
+    MVMObject *HOW = NULL, *meth = NULL;
 
-    MVMROOT(tc, obj, {
-        MVMROOT(tc, type, {
-            HOW = MVM_6model_get_how(tc, STABLE(type));
-            MVMROOT(tc, HOW, {
-                meth = MVM_6model_find_method_cache_only(tc, HOW,
-                    tc->instance->str_consts.accepts_type);
-            });
-        });
+    MVMROOT3(tc, obj, type, HOW, {
+        HOW = MVM_6model_get_how(tc, STABLE(type));
+        meth = MVM_6model_find_method_cache_only(tc, HOW,
+            tc->instance->str_consts.accepts_type);
     });
 
     if (!MVM_is_null(tc, meth)) {
@@ -321,7 +324,7 @@ static void accepts_type_sr(MVMThreadContext *tc, void *sr_data) {
 }
 
 static void mark_sr_data(MVMThreadContext *tc, MVMFrame *frame, MVMGCWorklist *worklist) {
-    AcceptsTypeSRData *atd = (AcceptsTypeSRData *)frame->special_return_data;
+    AcceptsTypeSRData *atd = (AcceptsTypeSRData *)frame->extra->special_return_data;
     MVM_gc_worklist_add(tc, worklist, &atd->obj);
     MVM_gc_worklist_add(tc, worklist, &atd->type);
 }
@@ -363,16 +366,12 @@ void MVM_6model_istype(MVMThreadContext *tc, MVMObject *obj, MVMObject *type, MV
     /* If we get here, need to call .^type_check on the value we're
      * checking, unless it's an accepts check. */
     if (!cache || (mode & MVM_TYPE_CHECK_CACHE_THEN_METHOD)) {
-        MVMObject *HOW, *meth;
+        MVMObject *HOW = NULL, *meth = NULL;
 
-        MVMROOT(tc, obj, {
-             MVMROOT(tc, type, {
-                 HOW = MVM_6model_get_how(tc, st);
-                 MVMROOT(tc, HOW, {
-                     meth = MVM_6model_find_method_cache_only(tc, HOW,
-                         tc->instance->str_consts.type_check);
-                 });
-             });
+        MVMROOT3(tc, obj, type, HOW, {
+            HOW = MVM_6model_get_how(tc, st);
+            meth = MVM_6model_find_method_cache_only(tc, HOW,
+                tc->instance->str_consts.type_check);
         });
         if (!MVM_is_null(tc, meth)) {
             /* Set up the call, using the result register as the target. */
@@ -388,9 +387,8 @@ void MVM_6model_istype(MVMThreadContext *tc, MVMObject *obj, MVMObject *type, MV
                 atd->obj = obj;
                 atd->type = type;
                 atd->res = res;
-                tc->cur_frame->special_return           = accepts_type_sr;
-                tc->cur_frame->special_return_data      = atd;
-                tc->cur_frame->mark_special_return_data = mark_sr_data;
+                MVM_frame_special_return(tc, tc->cur_frame, accepts_type_sr, NULL,
+                    atd, mark_sr_data);
             }
             STABLE(code)->invoke(tc, code, typecheck_callsite, tc->cur_frame->args);
             return;
@@ -450,7 +448,7 @@ MVMint64 MVM_6model_try_cache_type_check(MVMThreadContext *tc, MVMObject *obj, M
 
 /* Default invoke function on STables; for non-invokable objects */
 void MVM_6model_invoke_default(MVMThreadContext *tc, MVMObject *invokee, MVMCallsite *callsite, MVMRegister *args) {
-    MVM_exception_throw_adhoc(tc, "Cannot invoke this object (REPR: %s; %s)", REPR(invokee)->name, STABLE(invokee)->debug_name);
+    MVM_exception_throw_adhoc(tc, "Cannot invoke this object (REPR: %s; %s)", REPR(invokee)->name, MVM_6model_get_debug_name(tc, invokee));
 }
 
 /* Clean up STable memory. */
@@ -473,6 +471,24 @@ MVMuint64 MVM_6model_next_type_cache_id(MVMThreadContext *tc) {
     return (MVMuint64)MVM_add(&tc->instance->cur_type_cache_id, MVM_TYPE_CACHE_ID_INCR) + MVM_TYPE_CACHE_ID_INCR;
 }
 
+/* For type objects, marks the type as never repossessable. For concrete object
+ * instances, marks the individual ojbect as never repossessable. */
 void MVM_6model_never_repossess(MVMThreadContext *tc, MVMObject *obj) {
-    obj->header.flags |= MVM_CF_NEVER_REPOSSESS;
+    if (IS_CONCRETE(obj))
+        obj->header.flags |= MVM_CF_NEVER_REPOSSESS;
+    else
+        obj->st->mode_flags |= MVM_NEVER_REPOSSESS_TYPE;
+}
+
+/* Set the debug name on a type. */
+void MVM_6model_set_debug_name(MVMThreadContext *tc, MVMObject *type, MVMString *name) {
+    char *orig_debug_name;
+    uv_mutex_lock(&(tc->instance->mutex_free_at_safepoint));
+    orig_debug_name = STABLE(type)->debug_name;
+    if (orig_debug_name)
+        MVM_free_at_safepoint(tc, orig_debug_name);
+    STABLE(type)->debug_name = name && MVM_string_graphs(tc, name)
+        ? MVM_string_utf8_encode_C_string(tc, name)
+        : NULL;
+    uv_mutex_unlock(&(tc->instance->mutex_free_at_safepoint));
 }
