@@ -120,39 +120,36 @@ typedef enum {
     /* Has been promoted to the old generation. */
     MVM_CF_SECOND_GEN = 16,
 
-    /* Is shared - that is, more than one thread knows about it. */
-    MVM_CF_SHARED = 32,
-
     /* Has already been added to the gen2 aggregates pointing to nursery
      * objects list. */
-    MVM_CF_IN_GEN2_ROOT_LIST = 64,
+    MVM_CF_IN_GEN2_ROOT_LIST = 32,
 
     /* A full GC run has found this object to be live. */
-    MVM_CF_GEN2_LIVE = 128,
+    MVM_CF_GEN2_LIVE = 64,
 
     /* This object in fromspace is live with a valid forwarder. */
     /* TODO - should be possible to use the same bit for this and GEN2_LIVE. */
-    MVM_CF_FORWARDER_VALID = 256,
+    MVM_CF_FORWARDER_VALID = 128,
 
     /* Have we allocated memory to store a serialization index? */
-    MVM_CF_SERIALZATION_INDEX_ALLOCATED = 512,
+    MVM_CF_SERIALZATION_INDEX_ALLOCATED = 256,
 
     /* Have we arranged a persistent object ID for this object? */
-    MVM_CF_HAS_OBJECT_ID = 1024,
+    MVM_CF_HAS_OBJECT_ID = 512,
 
     /* Have we flagged this object as something we must never repossess? */
     /* Note: if you're hunting for a flag, some day in the future when we
      * have used them all, this one is easy enough to eliminate by having the
      * tiny number of objects marked this way in a remembered set. */
-    MVM_CF_NEVER_REPOSSESS = 2048,
+    MVM_CF_NEVER_REPOSSESS = 1024,
+
+    /* Is this object a nursery object that has been referenced from gen2?
+     * Used to promote it earlier. */
+    MVM_CF_REF_FROM_GEN2 = 2048,
 
     /* Has this item been chained into a gen2 freelist? This is only used in
      * GC debug more. */
     MVM_CF_DEBUG_IN_GEN2_FREE_LIST = 4096,
-
-    /* Is this object a nursery object that has been referenced from gen2?
-     * Used to promote it earlier. */
-    MVM_CF_REF_FROM_GEN2 = 8192,
 } MVMCollectableFlags;
 
 #ifdef MVM_USE_OVERFLOW_SERIALIZATION_INDEX
@@ -197,10 +194,8 @@ struct MVMCollectable {
         MVMSTable *st;
     } sc_forward_u;
 
-    /* Identifier of the thread that currently owns the object, if any. If the
-     * object is unshared, then this is always the creating thread. If it is
-     * shared then it's whoever currently holds the mutex on it, or 0 if there
-     * is no held mutex. */
+    /* Identifier of the thread that created the object. 0 if this is a
+     * non-heap frame. */
     MVMuint32 owner;
 
     /* Collectable flags (see MVMCollectableFlags). */
@@ -371,6 +366,10 @@ struct MVMSTable {
     /* If this STable is currently in the process of being repossessed. Used
      * to trigger clearup of memory pre-repossession. */
     MVMuint8 being_repossessed;
+
+    /* If this STable represents a type that can be the target of a
+     * change_type - that is to say, it's been mixed in to. */
+    MVMuint8 is_mixin_type;
 };
 
 /* The representation operations table. Note that representations are not
@@ -502,7 +501,7 @@ struct MVMREPROps_Positional {
      * they know the type of the passed array, otherwise they should use the REPR
      * API. */
     void (*splice) (MVMThreadContext *tc, MVMSTable *st,
-        MVMObject *root, void *data, MVMObject *target_array,
+        MVMObject *root, void *data, MVMObject *from,
         MVMint64 offset, MVMuint64 elems);
 
     /* Multi-dimensional array read. */
@@ -540,6 +539,13 @@ struct MVMREPROps_Positional {
     /* Multi-dim version of as_atomic. */
     AO_t * (*pos_as_atomic_multidim) (MVMThreadContext *tc, MVMSTable *st, MVMObject *root,
         void *data, MVMint64 num_indices, MVMint64 *indices);
+
+    void (*write_buf) (MVMThreadContext *tc, MVMSTable *st,
+        MVMObject *root, void *data, char *from,
+        MVMint64 offset, MVMuint64 elems);
+
+    MVMint64 (*read_buf) (MVMThreadContext *tc, MVMSTable *st,
+        MVMObject *root, void *data, MVMint64 offset, MVMuint64 elems);
 };
 struct MVMREPROps_Associative {
     /* Gets the value at the specified key and places it in the passed
