@@ -218,6 +218,10 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMSTable *st,
                     MVM_ASSIGN_REF(tc, &(st->header), repr_data->member_types[i], type);
                     if (inlined) {
                         MVMCArrayREPRData *carray_repr_data = (MVMCArrayREPRData *)STABLE(type)->REPR_data;
+                        if (!carray_repr_data) {
+                            MVM_exception_throw_adhoc(tc,
+                                "CPPStruct: can't inline a CArray attribute before its type's definition");
+                        }
                         bits                                = carray_repr_data->elem_size * 8;
                         repr_data->attribute_locations[i]  |= MVM_CSTRUCT_ATTR_INLINED;
 
@@ -245,9 +249,13 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMSTable *st,
                     MVM_ASSIGN_REF(tc, &(st->header), repr_data->member_types[i], type);
                     if (inlined) {
                         MVMCStructREPRData *cstruct_repr_data = (MVMCStructREPRData *)STABLE(type)->REPR_data;
+                        if (!cstruct_repr_data) {
+                            MVM_exception_throw_adhoc(tc,
+                                "CPPStruct: can't inline a CStruct attribute before its type's definition");
+                        }
                         bits                                  = cstruct_repr_data->struct_size * 8;
                         align                                 = cstruct_repr_data->struct_align;
-                        repr_data->attribute_locations[i]    |= MVM_CSTRUCT_ATTR_INLINED;
+                        repr_data->attribute_locations[i]    |= MVM_CPPSTRUCT_ATTR_INLINED;
                     }
                 }
                 else if (type_id == MVM_REPR_ID_MVMCPPStruct) {
@@ -257,6 +265,10 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMSTable *st,
                     MVM_ASSIGN_REF(tc, &(st->header), repr_data->member_types[i], type);
                     if (inlined) {
                         MVMCPPStructREPRData *cppstruct_repr_data = (MVMCPPStructREPRData *)STABLE(type)->REPR_data;
+                        if (!cppstruct_repr_data) {
+                            MVM_exception_throw_adhoc(tc,
+                                "CPPStruct: can't inline a CPPStruct attribute before its type's definition");
+                        }
                         bits                                      = cppstruct_repr_data->struct_size * 8;
                         align                                     = cppstruct_repr_data->struct_align;
                         repr_data->attribute_locations[i]        |= MVM_CPPSTRUCT_ATTR_INLINED;
@@ -265,13 +277,17 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMSTable *st,
                 else if (type_id == MVM_REPR_ID_MVMCUnion) {
                     /* It's a CUnion. */
                     repr_data->num_child_objs++;
-                    repr_data->attribute_locations[i] = (cur_obj_attr++ << MVM_CSTRUCT_ATTR_SHIFT) | MVM_CSTRUCT_ATTR_CUNION;
+                    repr_data->attribute_locations[i] = (cur_obj_attr++ << MVM_CPPSTRUCT_ATTR_SHIFT) | MVM_CPPSTRUCT_ATTR_CUNION;
                     MVM_ASSIGN_REF(tc, &(st->header), repr_data->member_types[i], type);
                     if (inlined) {
                         MVMCUnionREPRData *cunion_repr_data = (MVMCUnionREPRData *)STABLE(type)->REPR_data;
+                        if (!cunion_repr_data) {
+                            MVM_exception_throw_adhoc(tc,
+                                "CPPStruct: can't inline a CUnion attribute before its type's definition");
+                        }
                         bits                                = cunion_repr_data->struct_size * 8;
                         align                               = cunion_repr_data->struct_align;
-                        repr_data->attribute_locations[i]  |= MVM_CSTRUCT_ATTR_INLINED;
+                        repr_data->attribute_locations[i]  |= MVM_CPPSTRUCT_ATTR_INLINED;
                     }
                 }
                 else if (type_id == MVM_REPR_ID_MVMCPointer) {
@@ -404,7 +420,7 @@ static void copy_to(MVMThreadContext *tc, MVMSTable *st, void *src, MVMObject *d
 }
 
 /* Helper for complaining about attribute access errors. */
-MVM_NO_RETURN static void no_such_attribute(MVMThreadContext *tc, const char *action, MVMObject *class_handle, MVMString *name) MVM_NO_RETURN_GCC;
+MVM_NO_RETURN static void no_such_attribute(MVMThreadContext *tc, const char *action, MVMObject *class_handle, MVMString *name) MVM_NO_RETURN_ATTRIBUTE;
 static void no_such_attribute(MVMThreadContext *tc, const char *action, MVMObject *class_handle, MVMString *name) {
     char *c_name = MVM_string_utf8_encode_C_string(tc, name);
     char *waste[] = { c_name, NULL };
@@ -413,7 +429,7 @@ static void no_such_attribute(MVMThreadContext *tc, const char *action, MVMObjec
 }
 
 /* Helper to die because this type doesn't support attributes. */
-MVM_NO_RETURN static void die_no_attrs(MVMThreadContext *tc) MVM_NO_RETURN_GCC;
+MVM_NO_RETURN static void die_no_attrs(MVMThreadContext *tc) MVM_NO_RETURN_ATTRIBUTE;
 static void die_no_attrs(MVMThreadContext *tc) {
     MVM_exception_throw_adhoc(tc,
         "CPPStruct representation attribute not yet fully implemented");
@@ -483,7 +499,7 @@ static void get_attribute(MVMThreadContext *tc, MVMSTable *st, MVMObject *root,
                                 obj = MVM_repr_box_str(tc, typeobj, str);
                             });
                         }
-                        child_objs[real_slot] = obj;
+                        MVM_ASSIGN_REF(tc, &(root->header), body->child_objs[real_slot], obj);
                     }
                     else {
                         obj = typeobj;
