@@ -215,6 +215,10 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMObject *repr_in
                     MVM_ASSIGN_REF(tc, &(st->header), repr_data->member_types[i], type);
                     if (inlined) {
                         MVMCArrayREPRData *carray_repr_data = (MVMCArrayREPRData *)STABLE(type)->REPR_data;
+                        if (!carray_repr_data) {
+                            MVM_exception_throw_adhoc(tc,
+                                "CStruct: can't inline a CArray attribute before its type's definition");
+                        }
                         bits                                = carray_repr_data->elem_size * 8;
                         repr_data->attribute_locations[i]  |= MVM_CSTRUCT_ATTR_INLINED;
 
@@ -248,6 +252,10 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMObject *repr_in
                     MVM_ASSIGN_REF(tc, &(st->header), repr_data->member_types[i], type);
                     if (inlined) {
                         MVMCStructREPRData *cstruct_repr_data = (MVMCStructREPRData *)STABLE(type)->REPR_data;
+                        if (!cstruct_repr_data) {
+                            MVM_exception_throw_adhoc(tc,
+                                "CStruct: can't inline a CStruct attribute before its type's definition");
+                        }
                         bits                                  = cstruct_repr_data->struct_size * 8;
                         align                                 = cstruct_repr_data->struct_align;
                         repr_data->attribute_locations[i]    |= MVM_CSTRUCT_ATTR_INLINED;
@@ -260,6 +268,10 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMObject *repr_in
                     MVM_ASSIGN_REF(tc, &(st->header), repr_data->member_types[i], type);
                     if (inlined) {
                         MVMCPPStructREPRData *cppstruct_repr_data = (MVMCPPStructREPRData *)STABLE(type)->REPR_data;
+                        if (!cppstruct_repr_data) {
+                            MVM_exception_throw_adhoc(tc,
+                                "CStruct: can't inline a CPPStruct attribute before its type's definition");
+                        }
                         bits                                      = cppstruct_repr_data->struct_size * 8;
                         align                                     = cppstruct_repr_data->struct_align;
                         repr_data->attribute_locations[i]        |= MVM_CSTRUCT_ATTR_INLINED;
@@ -272,6 +284,10 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMObject *repr_in
                     MVM_ASSIGN_REF(tc, &(st->header), repr_data->member_types[i], type);
                     if (inlined) {
                         MVMCUnionREPRData *cunion_repr_data = (MVMCUnionREPRData *)STABLE(type)->REPR_data;
+                        if (!cunion_repr_data) {
+                            MVM_exception_throw_adhoc(tc,
+                                "CStruct: can't inline a CUnion attribute before its type's definition");
+                        }
                         bits                                = cunion_repr_data->struct_size * 8;
                         align                               = cunion_repr_data->struct_align;
                         repr_data->attribute_locations[i]  |= MVM_CSTRUCT_ATTR_INLINED;
@@ -497,7 +513,7 @@ static void get_attribute(MVMThreadContext *tc, MVMSTable *st, MVMObject *root,
                                 obj = MVM_repr_box_str(tc, typeobj, str);
                             });
                         }
-                        child_objs[real_slot] = obj;
+                        MVM_ASSIGN_REF(tc, &(root->header), body->child_objs[real_slot], obj);
                     }
                     else {
                         obj = typeobj;
@@ -748,6 +764,23 @@ static const MVMStorageSpec storage_spec = {
     0,                          /* is_unsigned */
 };
 
+static MVMuint64 unmanaged_size(MVMThreadContext *tc, MVMSTable *st, void *data) {
+    /* The CStruct data body itself is unmanaged, though it doesn't
+     * necessarily come from regular malloced heap memory */
+    MVMCStructREPRData *repr_data = (MVMCStructREPRData *)st->REPR_data;
+    MVMuint64 result = 0;
+
+    /* The allocated (or just-poisted-at) memory block */
+    /* TODO make sure when structs properly track "ownership" to
+     *      not add this for "child" structs */
+    result += repr_data->struct_size;
+
+    /* The array we hold wrapper objects in */
+    result += repr_data->num_child_objs * sizeof(MVMObject *);
+
+    return result;
+}
+
 /* Gets the storage specification for this representation. */
 static const MVMStorageSpec * get_storage_spec(MVMThreadContext *tc, MVMSTable *st) {
     return &storage_spec;
@@ -890,6 +923,8 @@ static void make_deref_op(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb
     MVM_spesh_usages_add_by_reg(tc, g, orig, deref_ins);
     MVM_spesh_get_facts(tc, g, deref_ins->operands[0])->writer = deref_ins;
 
+    MVM_spesh_graph_add_comment(tc, g, ins, "pointer dereference into CStruct body");
+
     MVM_spesh_manipulate_insert_ins(tc, bb, ins->prev, deref_ins);
 }
 static void spesh(MVMThreadContext *tc, MVMSTable *st, MVMSpeshGraph *g, MVMSpeshBB *bb, MVMSpeshIns *ins) {
@@ -901,7 +936,7 @@ static void spesh(MVMThreadContext *tc, MVMSTable *st, MVMSpeshGraph *g, MVMSpes
     if (!repr_data)
         return;
     switch (opcode) {
-    /*case MVM_OP_getattr_i:
+    case MVM_OP_getattr_i:
     case MVM_OP_getattrs_i: {
         MVMSpeshFacts *obj_facts = MVM_spesh_get_and_use_facts(tc, g, ins->operands[1]);
         MVMSpeshFacts *ch_facts = MVM_spesh_get_and_use_facts(tc, g, ins->operands[2]);
@@ -913,28 +948,33 @@ static void spesh(MVMThreadContext *tc, MVMSTable *st, MVMSpeshGraph *g, MVMSpes
                 MVMSTable      *flat_st = repr_data->flattened_stables[slot];
                 const MVMStorageSpec *flat_ss = flat_st->REPR->get_storage_spec(tc, flat_st);
                 add_slot_name_comment(tc, g, ins, name, ch_facts, st);
-                if (flat_st->REPR->ID == MVM_REPR_ID_P6int && (
-                            flat_ss->bits == 64 || flat_ss->bits == 32)) {
+                if (flat_st->REPR->ID == MVM_REPR_ID_P6int &&
+                        (flat_ss->bits == 64 || ((flat_ss->bits == 32 || flat_ss->bits == 16 || flat_ss->bits == 8)
+                        && !flat_ss->is_unsigned))) {
                     MVMSpeshOperand temp_reg = MVM_spesh_manipulate_get_temp_reg(tc, g, MVM_reg_int64);
                     MVMSpeshOperand orig_target = ins->operands[1];
-                    fprintf(stderr, "speshed a %s on CStruct. please don't explode\n", ins->info->name);
                     if (opcode == MVM_OP_getattrs_i)
                         MVM_spesh_usages_delete_by_reg(tc, g, ins->operands[3], ins);
                     MVM_spesh_usages_delete_by_reg(tc, g, ins->operands[2], ins);
+                    MVM_spesh_usages_delete_by_reg(tc, g, ins->operands[1], ins);
                     ins->info = MVM_op_get_op(
                             flat_ss->bits == 64 ? MVM_OP_sp_get_i64
-                                                : MVM_OP_sp_get_i32);
+                          : flat_ss->bits == 32 ? MVM_OP_sp_get_i32
+                          : flat_ss->bits == 16 ? MVM_OP_sp_get_i16
+                          :                       MVM_OP_sp_get_i8);
                     ins->operands[1] = temp_reg;
                     ins->operands[2].lit_i16 = repr_data->struct_offsets[slot];
 
                     MVM_spesh_usages_add_by_reg(tc, g, temp_reg, ins);
 
                     make_deref_op(tc, g, bb, ins, orig_target, temp_reg);
+
+                    MVM_spesh_manipulate_release_temp_reg(tc, g, temp_reg);
                 }
             }
         }
         break;
-    }*/
+    }
     case MVM_OP_bindattr_i:
     case MVM_OP_bindattrs_i: {
         MVMSpeshFacts *obj_facts = MVM_spesh_get_and_use_facts(tc, g, ins->operands[0]);
@@ -1018,6 +1058,6 @@ static const MVMREPROps CStruct_this_repr = {
     spesh, /* spesh */
     "CStruct", /* name */
     MVM_REPR_ID_MVMCStruct,
-    NULL, /* unmanaged_size */
+    unmanaged_size,
     NULL, /* describe_refs */
 };
