@@ -146,7 +146,60 @@ static void flip_return(MVMThreadContext *tc, void *sr_data) {
     MVMRegister *r = (MVMRegister *)sr_data;
     r->i64 = r->i64 ? 0 : 1;
 }
+/* ui64toa and i64toa
+ * Copyright(c) 2014-2016 Milo Yip (miloyip@gmail.com)
+ * https://github.com/miloyip/itoa-benchmark
+ * With minor modifications.
 
+ Copyright (C) 2014 Milo Yip
+
+MIT License:
+
+Permission is hereby granted, free of charge, to any person obtaining a copy
+of this software and associated documentation files (the "Software"), to deal
+in the Software without restriction, including without limitation the rights
+to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
+copies of the Software, and to permit persons to whom the Software is
+furnished to do so, subject to the following conditions:
+
+The above copyright notice and this permission notice shall be included in
+all copies or substantial portions of the Software.
+
+THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
+IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
+AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
+LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
+OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
+    THE SOFTWARE. */
+static char * u64toa_naive_worker(uint64_t value, char* buffer) {
+    char temp[20];
+    char *p = temp;
+    do {
+        *p++ = (char)(value % 10) + '0';
+        value /= 10;
+    } while (value > 0);
+
+    do {
+        *buffer++ = *--p;
+    } while (p != temp);
+
+    *buffer = '\0';
+    return buffer;
+}
+static size_t i64toa_naive(int64_t value, char* buffer) {
+    uint64_t u = value;
+    char *buf = buffer;
+    if (value < 0) {
+        *buf++ = '-';
+        u = ~u + 1;
+    }
+    return u64toa_naive_worker(u, buf) - buffer;
+}
+/* End code */
+static size_t u64toa_naive(uint64_t value, char* buffer) {
+    return u64toa_naive_worker(value, buffer) - buffer;
+}
 MVMString * MVM_coerce_i_s(MVMThreadContext *tc, MVMint64 i) {
     char buffer[64];
     int len;
@@ -157,11 +210,39 @@ MVMString * MVM_coerce_i_s(MVMThreadContext *tc, MVMint64 i) {
         if (cached)
             return cached;
     }
-
     /* Otherwise, need to do the work; cache it if in range. */
-    len = snprintf(buffer, 64, "%"PRIi64"", i);
+    len = i64toa_naive(i, buffer);
     if (0 <= len) {
-        MVMString *result = MVM_string_ascii_decode(tc, tc->instance->VMString, buffer, len);
+        MVMString *result = NULL;
+        MVMGrapheme8 *blob = MVM_malloc(len);
+        memcpy(blob, buffer, len);
+        result = MVM_string_ascii_from_buf_nocheck(tc, blob, len);
+        if (cache)
+            tc->instance->int_to_str_cache[i] = result;
+        return result;
+    }
+    else {
+        MVM_exception_throw_adhoc(tc, "Could not stringify integer");
+    }
+}
+
+MVMString * MVM_coerce_u_s(MVMThreadContext *tc, MVMuint64 i) {
+    char buffer[64];
+    int len;
+    /* See if we can hit the cache. */
+    int cache = i < MVM_INT_TO_STR_CACHE_SIZE;
+    if (cache) {
+        MVMString *cached = tc->instance->int_to_str_cache[i];
+        if (cached)
+            return cached;
+    }
+    /* Otherwise, need to do the work; cache it if in range. */
+    len = u64toa_naive(i, buffer);
+    if (0 <= len) {
+        MVMString *result = NULL;
+        MVMGrapheme8 *blob = MVM_malloc(len);
+        memcpy(blob, buffer, len);
+        result = MVM_string_ascii_from_buf_nocheck(tc, blob, len);
         if (cache)
             tc->instance->int_to_str_cache[i] = result;
         return result;
@@ -183,10 +264,14 @@ MVMString * MVM_coerce_n_s(MVMThreadContext *tc, MVMnum64 n) {
     }
     else {
         char buf[64];
-        int i;
         if (dtoa_grisu3(n, buf, 64) < 0)
             MVM_exception_throw_adhoc(tc, "Could not stringify number");
-        return MVM_string_ascii_decode(tc, tc->instance->VMString, buf, strlen(buf));
+        else {
+            MVMStringIndex len = strlen(buf);
+            MVMGrapheme8 *blob = MVM_malloc(len);
+            memcpy(blob, buf, len);
+            return MVM_string_ascii_from_buf_nocheck(tc, blob, len);
+        }
     }
 }
 
@@ -236,7 +321,7 @@ void MVM_coerce_smart_stringify(MVMThreadContext *tc, MVMObject *obj, MVMRegiste
         else if (ss->can_box & MVM_STORAGE_SPEC_CAN_BOX_NUM)
             res_reg->s = MVM_coerce_n_s(tc, REPR(obj)->box_funcs.get_num(tc, STABLE(obj), obj, OBJECT_BODY(obj)));
         else
-            MVM_exception_throw_adhoc(tc, "cannot stringify this");
+            MVM_exception_throw_adhoc(tc, "Cannot stringify this type (%s)", MVM_6model_get_stable_debug_name(tc, STABLE(obj)));
     }
 }
 
@@ -291,7 +376,54 @@ void MVM_coerce_smart_numify(MVMThreadContext *tc, MVMObject *obj, MVMRegister *
         else if (REPR(obj)->ID == MVM_REPR_ID_MVMHash)
             res_reg->n64 = (MVMnum64)REPR(obj)->elems(tc, STABLE(obj), obj, OBJECT_BODY(obj));
         else
-            MVM_exception_throw_adhoc(tc, "cannot numify this");
+            MVM_exception_throw_adhoc(tc, "Cannot numify this type (%s)", MVM_6model_get_stable_debug_name(tc, STABLE(obj)));
+    }
+}
+
+void MVM_coerce_smart_intify(MVMThreadContext *tc, MVMObject *obj, MVMRegister *res_reg) {
+    MVMObject *intmeth;
+
+    /* Handle null case. */
+    if (MVM_is_null(tc, obj)) {
+        res_reg->i64 = 0;
+    }
+
+    /* Check if there is an Int method. */
+    MVMROOT(tc, obj, {
+        intmeth = MVM_6model_find_method_cache_only(tc, obj,
+            tc->instance->str_consts.Int);
+    });
+
+    if (!MVM_is_null(tc, intmeth)) {
+        /* We need to do the invocation; just set it up with our result reg as
+         * the one for the call. */
+        MVMObject *code = MVM_frame_find_invokee(tc, intmeth, NULL);
+        MVMCallsite *inv_arg_callsite = MVM_callsite_get_common(tc, MVM_CALLSITE_ID_INV_ARG);
+
+        MVM_args_setup_thunk(tc, res_reg, MVM_RETURN_INT, inv_arg_callsite);
+        tc->cur_frame->args[0].o = obj;
+        STABLE(code)->invoke(tc, code, inv_arg_callsite, tc->cur_frame->args);
+        return;
+    }
+
+    /* Otherwise, guess something appropriate. */
+    if (!IS_CONCRETE(obj)) {
+        res_reg->i64 = 0;
+    }
+    else {
+        const MVMStorageSpec *ss = REPR(obj)->get_storage_spec(tc, STABLE(obj));
+        if (ss->can_box & MVM_STORAGE_SPEC_CAN_BOX_INT)
+            res_reg->i64 = REPR(obj)->box_funcs.get_int(tc, STABLE(obj), obj, OBJECT_BODY(obj));
+        else if (ss->can_box & MVM_STORAGE_SPEC_CAN_BOX_NUM)
+            res_reg->i64 = (MVMint64)REPR(obj)->box_funcs.get_num(tc, STABLE(obj), obj, OBJECT_BODY(obj));
+        else if (ss->can_box & MVM_STORAGE_SPEC_CAN_BOX_STR)
+            res_reg->i64 = MVM_coerce_s_i(tc, REPR(obj)->box_funcs.get_str(tc, STABLE(obj), obj, OBJECT_BODY(obj)));
+        else if (REPR(obj)->ID == MVM_REPR_ID_VMArray)
+            res_reg->i64 = REPR(obj)->elems(tc, STABLE(obj), obj, OBJECT_BODY(obj));
+        else if (REPR(obj)->ID == MVM_REPR_ID_MVMHash)
+            res_reg->i64 = REPR(obj)->elems(tc, STABLE(obj), obj, OBJECT_BODY(obj));
+        else
+            MVM_exception_throw_adhoc(tc, "Cannot intify this type (%s)", MVM_6model_get_stable_debug_name(tc, STABLE(obj)));
     }
 }
 
@@ -315,7 +447,7 @@ MVMint64 MVM_coerce_simple_intify(MVMThreadContext *tc, MVMObject *obj) {
         else if (REPR(obj)->ID == MVM_REPR_ID_MVMHash)
             return REPR(obj)->elems(tc, STABLE(obj), obj, OBJECT_BODY(obj));
         else
-            MVM_exception_throw_adhoc(tc, "cannot intify this");
+            MVM_exception_throw_adhoc(tc, "Cannot intify this type (%s)", MVM_6model_get_stable_debug_name(tc, STABLE(obj)));
     }
 }
 
@@ -420,17 +552,25 @@ void MVM_box_num(MVMThreadContext *tc, MVMnum64 value, MVMObject *type,
 
 }
 
+MVMString * MVM_unbox_str(MVMThreadContext *tc, MVMObject *obj) {
+    if (!IS_CONCRETE(obj))
+        MVM_exception_throw_adhoc(tc, "Cannot unbox a type object (%s) to a str.",
+            MVM_6model_get_debug_name(tc, obj));
+    return REPR(obj)->box_funcs.get_str(tc,
+        STABLE(obj), obj, OBJECT_BODY(obj));
+}
+
 void MVM_box_str(MVMThreadContext *tc, MVMString *value, MVMObject *type,
-                 MVMRegister * dst) {
+                 MVMRegister *dst) {
     MVMObject *box;
     MVMROOT(tc, value, {
-            box = REPR(type)->allocate(tc, STABLE(type));
-            if (REPR(box)->initialize)
-                REPR(box)->initialize(tc, STABLE(box), box, OBJECT_BODY(box));
-            REPR(box)->box_funcs.set_str(tc, STABLE(box), box,
-                                         OBJECT_BODY(box), value);
-            dst->o = box;
-        });
+        box = REPR(type)->allocate(tc, STABLE(type));
+        if (REPR(box)->initialize)
+            REPR(box)->initialize(tc, STABLE(box), box, OBJECT_BODY(box));
+        REPR(box)->box_funcs.set_str(tc, STABLE(box), box,
+                                     OBJECT_BODY(box), value);
+        dst->o = box;
+    });
 }
 
 void MVM_box_uint(MVMThreadContext *tc, MVMuint64 value, MVMObject *type,
