@@ -82,8 +82,8 @@ typedef struct {
 static MVMint32 handler_can_handle(MVMFrame *f, MVMFrameHandler *fh, MVMint32 cat, MVMObject *payload) {
     MVMuint32         category_mask = fh->category_mask;
     MVMuint64       block_has_label = category_mask & MVM_EX_CAT_LABELED;
-    MVMuint64           block_label = block_has_label ? (MVMuint64)(f->work[fh->label_reg].o) : 0;
-    MVMuint64          thrown_label = payload ? (MVMuint64)payload : 0;
+    MVMuint64           block_label = block_has_label ? (uintptr_t)(f->work[fh->label_reg].o) : 0;
+    MVMuint64          thrown_label = payload ? (uintptr_t)payload : 0;
     MVMuint64 identical_label_found = thrown_label == block_label;
     return ((cat & category_mask) == cat && (!(cat & MVM_EX_CAT_LABELED) || identical_label_found))
         || ((category_mask & MVM_EX_CAT_CONTROL) && cat != MVM_EX_CAT_CATCH);
@@ -360,6 +360,7 @@ static void run_handler(MVMThreadContext *tc, LocatedHandler lh, MVMObject *ex_o
         /* Create active handler record. */
         MVMActiveHandler *ah = MVM_malloc(sizeof(MVMActiveHandler));
         MVMFrame *cur_frame = tc->cur_frame;
+        MVMFrame *pres_frame;
         MVMObject *handler_code;
 
         /* Ensure we have an exception object. */
@@ -371,6 +372,14 @@ static void run_handler(MVMThreadContext *tc, LocatedHandler lh, MVMObject *ex_o
             MVM_ASSIGN_REF(tc, &(ex_obj->header), ((MVMException *)ex_obj)->body.payload, payload);
         }
 
+        /* Preserve the frame caller chain, so we can do backtraces. */
+        pres_frame = ((MVMException *)ex_obj)->body.origin;
+        while (pres_frame) {
+            MVMFrameExtra *extra = MVM_frame_extra(tc, pres_frame);
+            extra->caller_info_needed = 1;
+            pres_frame = pres_frame->caller;
+        }
+
         /* Find frame to invoke. */
         handler_code = MVM_frame_find_invokee(tc, lh.frame->work[lh.handler->block_reg].o, NULL);
 
