@@ -33,6 +33,7 @@ typedef struct {
 /* Various common string constants. */
 struct MVMStringConsts {
     MVMString *empty;
+    MVMString *Int;
     MVMString *Str;
     MVMString *Num;
     MVMString *integer;
@@ -246,6 +247,12 @@ struct MVMInstance {
      * than per STable, as this doesn't happen often. */
     uv_mutex_t mutex_parameterization_add;
 
+    /************************************************************************
+     * Configuration programs (for spesh, profiler, ...)
+     ************************************************************************/
+
+    MVMConfigurationProgram *confprog;
+
     /************************************************************************
      * Specializer (dynamic optimization)
      ************************************************************************/
@@ -258,6 +265,7 @@ struct MVMInstance {
     MVMint8 spesh_inline_enabled;
     MVMint8 spesh_inline_log;
     MVMint8 spesh_osr_enabled;
+    MVMint8 spesh_pea_enabled;
     MVMint8 spesh_nodelay;
     MVMint8 spesh_blocking;
 
@@ -269,6 +277,9 @@ struct MVMInstance {
     /* Mutex taken when install specializations. */
     uv_mutex_t mutex_spesh_install;
 
+    /* The thread object representing the spesh thread */
+    MVMObject *spesh_thread;
+
     /* The concurrent queue used to send logs to spesh_thread, provided it
      * is enabled. */
     MVMObject *spesh_queue;
@@ -293,24 +304,21 @@ struct MVMInstance {
      ************************************************************************/
 
     /* Flag for if jit is enabled */
-    MVMint32 jit_enabled;
-
-    MVMint32 jit_expr_enabled;
+    MVMuint8 jit_enabled;
+    MVMuint8 jit_expr_enabled;
+    MVMuint8 jit_debug_enabled;
 
     /* bisection flags, to stop the JIT from using the expression compiler above
      * certain frame seq nr / basic blocks nrs, allowing a debugger to figure
      * out where a particular piece of code breaks */
     MVMint32 jit_expr_last_frame;
     MVMint32 jit_expr_last_bb;
-    /* File for JIT logging */
-    FILE *jit_log_fh;
+    /* File for JIT perf map logging */
+    FILE *jit_perf_map;
 
     /* Directory name for JIT bytecode dumps */
     char *jit_bytecode_dir;
 
-    /* File for map of frame information for bytecode dumps */
-    FILE *jit_bytecode_map;
-
     /* sequence number for JIT compiled frames */
     MVMint32 jit_seq_nr;
 
@@ -328,13 +336,14 @@ struct MVMInstance {
      * queue of tasks that need to be processed by the event loop thread
      * and an array of active tasks, for the purpose of keeping them GC
      * marked. */
-    MVMThreadContext *event_loop_thread;
-    uv_mutex_t        mutex_event_loop_start;
-    uv_sem_t          sem_event_loop_started;
+    MVMObject        *event_loop_thread;
+    uv_loop_t        *event_loop;
+    uv_mutex_t        mutex_event_loop;
     MVMObject        *event_loop_todo_queue;
     MVMObject        *event_loop_permit_queue;
     MVMObject        *event_loop_cancel_queue;
     MVMObject        *event_loop_active;
+    MVMObject        *event_loop_free_indices;
     uv_async_t       *event_loop_wakeup;
 
     /* Standard file handles. */
@@ -500,6 +509,9 @@ struct MVMInstance {
     FILE *coverage_log_fh;
     MVMuint32  coverage_control;
 
+    /* The time it takes to run the profiler instrumentation. */
+    MVMuint64 profiling_overhead;
+
     /************************************************************************
      * Debugging
      ************************************************************************/
