@@ -1,6 +1,9 @@
 #define MVMGCSTATUS_MASK 3
 #define MVMSUSPENDSTATUS_MASK 12
 
+#define MVM_NUM_TEMP_BIGINTS 3
+#include "tommath.h"
+
 /* Possible values for the thread execution interrupt flag. */
 typedef enum {
     /* Indicates that the thread is currently executing, and should
@@ -265,9 +268,6 @@ struct MVMThreadContext {
      * Per-thread state held by assorted VM subsystems
      ************************************************************************/
 
-    /* libuv event loop */
-    uv_loop_t *loop;
-
     /* Mutex that must be released if we throw an exception. Used in places
      * like I/O, which grab a mutex but may throw an exception. */
     uv_mutex_t *ex_release_mutex;
@@ -298,6 +298,10 @@ struct MVMThreadContext {
     /* Random number generator state. */
     MVMuint64 rand_state[2];
 
+    /* Temporary big integers for when we need to upgrade operands in order
+     * to perform an operation. */
+    mp_int *temp_bigints[MVM_NUM_TEMP_BIGINTS];
+
     /* NFA evaluator memory cache, to avoid many allocations; see NFA.c. */
     MVMuint32 *nfa_done;
     MVMuint32 *nfa_curst;
