@@ -1,5 +1,6 @@
 #include "moar.h"
 #include <platform/threads.h>
+#include "platform/malloc_trim.h"
 
 /* If we have the job of doing GC for a thread, we add it to our work
  * list. */
@@ -45,6 +46,7 @@ static MVMuint32 signal_one_thread(MVMThreadContext *tc, MVMThreadContext *to_si
             case MVMGCStatus_UNABLE | MVMSuspendState_SUSPEND_REQUEST:
             case MVMGCStatus_UNABLE | MVMSuspendState_SUSPENDED:
                 had_suspend_request = current & MVMSUSPENDSTATUS_MASK;
+                /* fallthrough */
             case MVMGCStatus_UNABLE:
                 /* Otherwise, it's blocked; try to set it to work Stolen. */
                 if (MVM_cas(&to_signal->gc_status, MVMGCStatus_UNABLE | had_suspend_request,
@@ -183,10 +185,7 @@ static void finish_gc(MVMThreadContext *tc, MVMuint8 gen, MVMuint8 is_coordinato
         GCDEBUG_LOG(tc, MVM_GC_DEBUG_ORCHESTRATE,
             "Thread %d run %d : Co-ordinator handling fixed-size allocator safepoint frees\n");
         MVM_fixed_size_safepoint(tc, tc->instance->fsa);
-
-        MVM_profile_dump_instrumented_data(tc);
-        MVM_profile_heap_take_snapshot(tc);
-
+        MVM_alloc_safepoint(tc);
         GCDEBUG_LOG(tc, MVM_GC_DEBUG_ORCHESTRATE,
             "Thread %d run %d : Co-ordinator signalling in-trays clear\n");
         uv_mutex_lock(&tc->instance->mutex_gc_orchestrate);
@@ -223,10 +222,13 @@ static void finish_gc(MVMThreadContext *tc, MVMuint8 gen, MVMuint8 is_coordinato
         else {
             /* Free gen2 unmarked if full collection. */
             if (gen == MVMGCGenerations_Both) {
+                /* Tell malloc implementation to free empty pages to kernel.
+                 * Currently only activated for Linux. */
+                MVM_malloc_trim();
                 GCDEBUG_LOG(tc, MVM_GC_DEBUG_ORCHESTRATE,
                     "Thread %d run %d : freeing gen2 of thread %d\n",
                     other->thread_id);
-                MVM_gc_collect_free_gen2_unmarked(other, 0);
+                MVM_gc_collect_free_gen2_unmarked(tc, other, 0);
             }
 
             /* Contribute this thread's promoted bytes. */
@@ -236,7 +238,7 @@ static void finish_gc(MVMThreadContext *tc, MVMuint8 gen, MVMuint8 is_coordinato
             GCDEBUG_LOG(tc, MVM_GC_DEBUG_ORCHESTRATE,
                 "Thread %d run %d : collecting nursery uncopied of thread %d\n",
                 other->thread_id);
-            MVM_gc_collect_free_nursery_uncopied(other, tc->gc_work[i].limit);
+            MVM_gc_collect_free_nursery_uncopied(tc, other, tc->gc_work[i].limit);
 
             /* Handle exited threads. */
             if (MVM_load(&thread_obj->body.stage) == MVM_thread_stage_exited) {
@@ -253,6 +255,11 @@ static void finish_gc(MVMThreadContext *tc, MVMuint8 gen, MVMuint8 is_coordinato
         }
     }
 
+    if (is_coordinator) {
+        MVM_profile_dump_instrumented_data(tc);
+        MVM_profile_heap_take_snapshot(tc);
+    }
+
     /* Signal acknowledgement of completing the cleanup,
      * except for STables, and if we're the final to do
      * so, free the STables, which have been linked. */
@@ -667,8 +674,8 @@ void MVM_gc_global_destruction(MVMThreadContext *tc) {
     tc->nursery_tospace = nursery_tmp;
 
     /* Run the objects' finalizers */
-    MVM_gc_collect_free_nursery_uncopied(tc, tc->nursery_alloc);
+    MVM_gc_collect_free_nursery_uncopied(tc, tc, tc->nursery_alloc);
     MVM_gc_root_gen2_cleanup(tc);
-    MVM_gc_collect_free_gen2_unmarked(tc, 1);
+    MVM_gc_collect_free_gen2_unmarked(tc, tc, 1);
     MVM_gc_collect_free_stables(tc);
 }
