@@ -74,16 +74,25 @@ void MVM_gc_root_add_instance_roots_to_worklist(MVMThreadContext *tc, MVMGCWorkl
     add_collectable(tc, worklist, snapshot, tc->instance->compiler_registry, "Compiler registry");
     add_collectable(tc, worklist, snapshot, tc->instance->hll_syms, "HLL symbols");
     add_collectable(tc, worklist, snapshot, tc->instance->clargs, "Command line args");
+
+    add_collectable(tc, worklist, snapshot, tc->instance->event_loop_thread,
+        "Event loop thread");
     add_collectable(tc, worklist, snapshot, tc->instance->event_loop_todo_queue,
         "Event loop todo queue");
     add_collectable(tc, worklist, snapshot, tc->instance->event_loop_permit_queue,
         "Event loop permit queue");
     add_collectable(tc, worklist, snapshot, tc->instance->event_loop_cancel_queue,
         "Event loop cancel queue");
-    add_collectable(tc, worklist, snapshot, tc->instance->event_loop_active, "Event loop active");
+    add_collectable(tc, worklist, snapshot, tc->instance->event_loop_active,
+        "Event loop active task list");
+    add_collectable(tc, worklist, snapshot, tc->instance->event_loop_free_indices,
+        "Event loop active free indices list");
 
+    add_collectable(tc, worklist, snapshot, tc->instance->spesh_thread,
+        "Specialization thread");
     add_collectable(tc, worklist, snapshot, tc->instance->spesh_queue,
         "Specialization log queue");
+
     if (worklist)
         MVM_spesh_plan_gc_mark(tc, tc->instance->spesh_plan, worklist);
 
@@ -122,6 +131,9 @@ void MVM_gc_root_add_instance_roots_to_worklist(MVMThreadContext *tc, MVMGCWorkl
     add_collectable(tc, worklist, snapshot, tc->instance->sig_arr,
         "Cached signal mapping array");
 
+    if (tc->instance->confprog)
+        MVM_confprog_mark(tc, worklist, snapshot);
+
     MVM_debugserver_mark_handles(tc, worklist, snapshot);
 }
 
@@ -191,8 +203,12 @@ void MVM_gc_root_add_tc_roots_to_worklist(MVMThreadContext *tc, MVMGCWorklist *w
         MVM_spesh_sim_stack_gc_mark(tc, tc->spesh_sim_stack, worklist);
     else
         MVM_spesh_sim_stack_gc_describe(tc, snapshot, tc->spesh_sim_stack);
-    if (tc->spesh_active_graph)
-        MVM_spesh_graph_mark(tc, tc->spesh_active_graph, worklist);
+    if (tc->spesh_active_graph) {
+        if (worklist)
+            MVM_spesh_graph_mark(tc, tc->spesh_active_graph, worklist);
+        else
+            MVM_spesh_graph_describe(tc, tc->spesh_active_graph, snapshot);
+    }
     MVM_spesh_plugin_guard_list_mark(tc, tc->plugin_guards, tc->num_plugin_guards, worklist);
     add_collectable(tc, worklist, snapshot, tc->plugin_guard_args,
         "Plugin guard args");
