@@ -148,6 +148,7 @@ MVMint64 socket_read_bytes(MVMThreadContext *tc, MVMOSHandle *h, char **buf, MVM
             /* Still something left in the just-read packet for next time. */
             data->last_packet_start += bytes - last_available;
         }
+        MVM_free(use_last_packet);
     }
     else if (data->last_packet) {
         /* Only data from the just-read packet. */
@@ -171,6 +172,7 @@ MVMint64 socket_read_bytes(MVMThreadContext *tc, MVMOSHandle *h, char **buf, MVM
         *buf = MVM_malloc(bytes);
         memcpy(*buf, use_last_packet + use_last_packet_start, bytes);
         data->eof = 1;
+        MVM_free(use_last_packet);
     }
     else {
         /* Nothing to hand back; at EOF. */
@@ -439,7 +441,7 @@ MVMint64 socket_getport(MVMThreadContext *tc, MVMOSHandle *h) {
 
 static MVMint64 socket_is_tty(MVMThreadContext *tc, MVMOSHandle *h) {
     MVMIOSyncSocketData *data = (MVMIOSyncSocketData *)h->body.data;
-    return isatty(data->handle);
+    return (MVMint64)isatty(data->handle);
 }
 
 static MVMint64 socket_handle(MVMThreadContext *tc, MVMOSHandle *h) {
@@ -515,7 +517,13 @@ MVMObject * MVM_io_socket_create(MVMThreadContext *tc, MVMint64 listen) {
 }
 
 MVMString * MVM_io_get_hostname(MVMThreadContext *tc) {
-    char hostname[65];
-    gethostname(hostname, 65);
+    char hostname[UV_MAXHOSTNAMESIZE];
+    size_t size = UV_MAXHOSTNAMESIZE;
+    int result = uv_os_gethostname(hostname, &size);
+
+    if(result < 0) {
+        MVM_exception_throw_adhoc(tc, "Failed to get hostname: %i", result);
+    }
+
     return MVM_string_ascii_decode_nt(tc, tc->instance->VMString, hostname);
 }
