@@ -110,6 +110,7 @@ typedef enum {
        not known yet, i.e. for ahead of time compiled native calls. */
     MVM_JIT_ARG_I64,
     MVM_JIT_ARG_I64_RW,
+    MVM_JIT_ARG_DOUBLE,
     /* Pointers are passed as objects with CPointer representation, i.e. the
        actual pointer is part of the object's data. The MVM_JIT_ARG_PTR type
        unboxes the CPointer object and passes on the contained pointer */
@@ -120,6 +121,7 @@ typedef enum {
        is already known. */
     MVM_JIT_PARAM_I64,
     MVM_JIT_PARAM_I64_RW,
+    MVM_JIT_PARAM_DOUBLE,
     MVM_JIT_PARAM_PTR,
     MVM_JIT_PARAM_VMARRAY,
     /* spesh slot value */
@@ -170,7 +172,6 @@ struct MVMJitCallC {
     void       *func_ptr;
     MVMJitCallArg  *args;
     MVMuint16   num_args;
-    MVMuint16  has_vargs;
     MVMJitRVMode rv_mode;
     MVMint16      rv_idx;
 };
