@@ -85,14 +85,16 @@ The following environment variables are respected:\n\
     MVM_SPESH_DISABLE           Disables all dynamic optimization\n\
     MVM_SPESH_INLINE_DISABLE    Disables inlining\n\
     MVM_SPESH_OSR_DISABLE       Disables on-stack replacement\n\
+    MVM_SPESH_PEA_DISABLE       Disables partial escape analysis and related optimizations\n\
     MVM_SPESH_BLOCKING          Blocks log-sending thread while specializer runs\n\
     MVM_SPESH_LOG               Specifies a dynamic optimizer log file\n\
     MVM_SPESH_NODELAY           Run dynamic optimization even for cold frames\n\
     MVM_SPESH_LIMIT             Limit the maximum number of specializations\n\
     MVM_JIT_DISABLE             Disables JITting to machine code\n\
     MVM_JIT_EXPR_DISABLE        Disable advanced 'expression' JIT\n\
-    MVM_JIT_LOG                 Specifies a JIT-compiler log file\n\
-    MVM_JIT_BYTECODE_DIR        Specifies a directory for JIT bytecode dumps\n\
+    MVM_JIT_DEBUG               Add JIT debugging information to spesh log\n\
+    MVM_JIT_PERF_MAP            Create a map file for the 'perf' profiler (linux only)\n\
+    MVM_JIT_DUMP_BYTECODE       Dump bytecode in temporary directory\n\
     MVM_SPESH_INLINE_LOG        Dump details of inlining attempts to stderr\n\
     MVM_CROSS_THREAD_WRITE_LOG  Log unprotected cross-thread object writes to stderr\n\
     MVM_COVERAGE_LOG            Append (de-duped by default) line-by-line coverage messages to this file\n\
@@ -224,15 +226,15 @@ int wmain(int argc, wchar_t *wargv[])
 
             case OPT_DEBUGPORT: {
                 MVMint64 port;
-                char *portstr = argv[argi] + strlen("--debugport=") + 1;
+                char *portstr = argv[argi] + strlen("--debug-port=");
                 char *endptr;
                 port = strtoll(portstr, &endptr, 10);
                 if (*endptr != '\0') {
                     fprintf(stderr, "ERROR: Invalid characters in debug port flag: %s\n", portstr);
                     return EXIT_FAILURE;
                 }
-                if (port <= 1024 || port > 65535) {
-                    fprintf(stderr, "ERROR: debug server port out of range. We only accept ports above 1024 and below 65535. (got: %lu)\n", port);
+                if (port <= 1024 || 65535 < port) {
+                    fprintf(stderr, "ERROR: debug server port out of range. We only accept ports above 1024 and below 65535. (got: %"PRIi64")\n", port);
                     return EXIT_FAILURE;
                 }
                 debugserverport = (MVMuint32)port;
