@@ -211,6 +211,7 @@ MVM_PUBLIC const MVMint32 MVM_jit_support(void);
 #include "profiler/profile.h"
 #include "profiler/heapsnapshot.h"
 #include "profiler/telemeh.h"
+#include "profiler/configuration.h"
 #include "instrument/crossthreadwrite.h"
 #include "instrument/line_coverage.h"
 
@@ -227,6 +228,9 @@ MVM_PUBLIC void MVM_vm_set_exec_name(MVMInstance *instance, const char *exec_nam
 MVM_PUBLIC void MVM_vm_set_prog_name(MVMInstance *instance, const char *prog_name);
 MVM_PUBLIC void MVM_vm_set_lib_path(MVMInstance *instance, int count, const char **lib_path);
 
+/* Returns absolute executable path. */
+MVM_PUBLIC int MVM_exepath(char* buffer, size_t* size);
+
 #if defined(__s390__)
 AO_t AO_fetch_compare_and_swap_emulation(volatile AO_t *addr, AO_t old_val, AO_t new_val);
 # define AO_fetch_compare_and_swap_full(addr, old, newval) \
