@@ -92,8 +92,8 @@ static void decont_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshIns *in
      * info. */
     MVMint32 in_flags = in_facts->flags;
     if ((in_flags & MVM_SPESH_FACT_TYPEOBJ) ||
-            (in_flags & MVM_SPESH_FACT_KNOWN_TYPE) &&
-            !in_facts->type->st->container_spec) {
+            ((in_flags & MVM_SPESH_FACT_KNOWN_TYPE) &&
+            !in_facts->type->st->container_spec)) {
         copy_facts(tc, g, out_orig, out_i, in_orig, in_i);
         return;
     }
@@ -515,6 +515,11 @@ static void add_bb_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb,
                 ins->operands[0].reg.orig, ins->operands[0].reg.i,
                 ins->operands[2].reg.orig, ins->operands[2].reg.i);
             break;
+        case MVM_OP_nativecallcast:
+            create_facts(tc, g,
+                ins->operands[0].reg.orig, ins->operands[0].reg.i,
+                ins->operands[2].reg.orig, ins->operands[2].reg.i);
+            break;
         case MVM_OP_trunc_u16:
         case MVM_OP_trunc_i16:
             trunc_i16_facts(tc, g, ins);
