@@ -63,6 +63,35 @@ void MVM_spesh_graph_add_deopt_annotation(MVMThreadContext *tc, MVMSpeshGraph *g
     g->num_deopt_addrs++;
 }
 
+MVM_FORMAT(printf, 4, 5)
+MVM_PUBLIC void MVM_spesh_graph_add_comment(MVMThreadContext *tc, MVMSpeshGraph *g,
+    MVMSpeshIns *ins, const char *fmt, ...) {
+    size_t size;
+    char *comment;
+    va_list ap;
+    MVMSpeshAnn *ann;
+
+    if (!MVM_spesh_debug_enabled(tc))
+        return;
+
+    va_start(ap, fmt);
+
+    size = vsnprintf(NULL, 0, fmt, ap);
+    comment = MVM_spesh_alloc(tc, g, ++size);
+
+    va_end(ap);
+
+    ann               = MVM_spesh_alloc(tc, g, sizeof(MVMSpeshAnn));
+    ann->type         = MVM_SPESH_ANN_COMMENT;
+    ann->data.comment = comment;
+    ann->next         = ins->annotations;
+    ins->annotations  = ann;
+
+    va_start(ap, fmt);
+    vsnprintf(comment, size, fmt, ap);
+    va_end(ap);
+}
+
 /* Records the current bytecode position as a logged annotation. Used for
  * resolving logged values. */
 static void add_logged_annotation(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshIns *ins_node,
@@ -105,7 +134,8 @@ static MVMint32 already_succs(MVMThreadContext *tc, MVMSpeshBB *bb, MVMSpeshBB *
 #define MVM_CFG_BB_END      2
 #define MVM_CFG_BB_JUMPLIST 4
 static void build_cfg(MVMThreadContext *tc, MVMSpeshGraph *g, MVMStaticFrame *sf,
-                      MVMint32 *existing_deopts, MVMint32 num_existing_deopts) {
+                      MVMint32 *existing_deopts, MVMint32 num_existing_deopts,
+                      MVMint32 *deopt_usage_info, MVMSpeshIns ***deopt_usage_ins_out) {
     MVMSpeshBB  *cur_bb, *prev_bb;
     MVMSpeshIns *last_ins;
     MVMint64     i;
@@ -379,8 +409,10 @@ static void build_cfg(MVMThreadContext *tc, MVMSpeshGraph *g, MVMStaticFrame *sf
     /* Annotate instructions that are handler-significant. */
     for (i = 0; i < g->num_handlers; i++) {
         /* Start or got may be -1 if the code the handler covered became
-         * dead. If so, mark the handler as removed. */
-        if (g->handlers[i].start_offset == -1 || g->handlers[i].goto_offset == -1) {
+         * dead. If so, mark the handler as removed. Ditto if end is
+         * before start (would never match). */
+        if (g->handlers[i].start_offset == -1 || g->handlers[i].goto_offset == -1 ||
+                g->handlers[i].start_offset > g->handlers[i].end_offset) {
             if (!g->unreachable_handlers)
                 g->unreachable_handlers = MVM_spesh_alloc(tc, g, g->num_handlers);
             g->unreachable_handlers[i] = 1;
@@ -544,6 +576,9 @@ static void build_cfg(MVMThreadContext *tc, MVMSpeshGraph *g, MVMStaticFrame *sf
              * mark it as an active handler. Unmark those where we see the
              * end of the handler. */
             if (cur_bb->first_ins->annotations) {
+                /* Process them in two passes in case we have two on the
+                 * same instruction and disordered. */
+                MVMuint32 has_end = 0;
                 MVMSpeshAnn *ann = cur_bb->first_ins->annotations;
                 while (ann) {
                     switch (ann->type) {
@@ -554,14 +589,25 @@ static void build_cfg(MVMThreadContext *tc, MVMSpeshGraph *g, MVMStaticFrame *sf
                             }
                             break;
                         case MVM_SPESH_ANN_FH_END:
-                            if (!is_catch_handler(tc, g, ann->data.frame_handler_index)) {
-                                active_handlers[ann->data.frame_handler_index] = 0;
-                                num_active_handlers--;
-                            }
+                            has_end = 1;
                             break;
                     }
                     ann = ann->next;
                 }
+                if (has_end) {
+                    ann = cur_bb->first_ins->annotations;
+                    while (ann) {
+                        switch (ann->type) {
+                            case MVM_SPESH_ANN_FH_END:
+                                if (!is_catch_handler(tc, g, ann->data.frame_handler_index)) {
+                                    active_handlers[ann->data.frame_handler_index] = 0;
+                                    num_active_handlers--;
+                                }
+                                break;
+                        }
+                        ann = ann->next;
+                    }
+                }
             }
 
             /* Consider the last instruction, to see how we leave the BB. */
@@ -669,6 +715,24 @@ static void build_cfg(MVMThreadContext *tc, MVMSpeshGraph *g, MVMStaticFrame *sf
         }
     }
 
+    /* We may also need to reconstruct deopt usage info, to allow optimizing of
+     * inlinees. That is persisted using bytecode offsets, and this is the last
+     * place we can map those into instructions, so do it here if needed. */
+    if (deopt_usage_ins_out && deopt_usage_info) {
+        MVM_VECTOR_DECL(MVMSpeshIns *, usage_ins);
+        MVMuint32 idx = 0;
+        MVM_VECTOR_INIT(usage_ins, 32);
+        while (1) {
+            MVMint32 offset = deopt_usage_info[idx];
+            if (offset == -1)
+                break;
+            MVM_VECTOR_PUSH(usage_ins, ins_flat[byte_to_ins_flags[offset] >> 3]);
+            idx++;
+            idx += deopt_usage_info[idx] + 1; /* Skip over deopt indices */
+        }
+        *deopt_usage_ins_out = usage_ins;
+    }
+
     /* Clear up the temporary arrays. */
     MVM_free(byte_to_ins_flags);
     MVM_free(ins_flat);
@@ -707,6 +771,7 @@ static void insert_object_null_instructions(MVMThreadContext *tc, MVMSpeshGraph
             null_ins->operands = MVM_spesh_alloc(tc, g, sizeof(MVMSpeshOperand));
             null_ins->operands[0].reg.orig = i;
             MVM_spesh_manipulate_insert_ins(tc, insert_bb, insert_after, null_ins);
+            insert_after = null_ins;
         }
     }
 }
@@ -744,7 +809,7 @@ static void dfs(MVMSpeshBB **rpo, MVMint32 *insert_pos, MVMuint8 *seen, MVMSpesh
     bb->rpo_idx = *insert_pos;
     (*insert_pos)--;
 }
-static MVMSpeshBB ** reverse_postorder(MVMThreadContext *tc, MVMSpeshGraph *g) {
+MVMSpeshBB ** MVM_spesh_graph_reverse_postorder(MVMThreadContext *tc, MVMSpeshGraph *g) {
     MVMSpeshBB **rpo  = MVM_calloc(g->num_bbs, sizeof(MVMSpeshBB *));
     MVMuint8    *seen = MVM_calloc(g->num_bbs, 1);
     MVMint32     ins  = g->num_bbs - 1;
@@ -1177,7 +1242,7 @@ static void ssa(MVMThreadContext *tc, MVMSpeshGraph *g) {
     MVMint32 i, num_locals;
 
     /* Compute dominance frontiers. */
-    MVMSpeshBB **rpo  = reverse_postorder(tc, g);
+    MVMSpeshBB **rpo  = MVM_spesh_graph_reverse_postorder(tc, g);
     MVMint32    *doms = compute_dominators(tc, g, rpo);
     add_children(tc, g, rpo, doms);
     add_dominance_frontiers(tc, g, rpo, doms);
@@ -1228,7 +1293,7 @@ MVMSpeshGraph * MVM_spesh_graph_create(MVMThreadContext *tc, MVMStaticFrame *sf,
     }
 
     /* Build the CFG out of the static frame, and transform it to SSA. */
-    build_cfg(tc, g, sf, NULL, 0);
+    build_cfg(tc, g, sf, NULL, 0, NULL, NULL);
     if (insert_object_nulls)
         insert_object_null_instructions(tc, g);
     if (!cfg_only) {
@@ -1243,7 +1308,8 @@ MVMSpeshGraph * MVM_spesh_graph_create(MVMThreadContext *tc, MVMStaticFrame *sf,
 
 /* Takes a static frame and creates a spesh graph for it. */
 MVMSpeshGraph * MVM_spesh_graph_create_from_cand(MVMThreadContext *tc, MVMStaticFrame *sf,
-                                                 MVMSpeshCandidate *cand, MVMuint32 cfg_only) {
+                                                 MVMSpeshCandidate *cand, MVMuint32 cfg_only,
+                                                 MVMSpeshIns ***deopt_usage_ins_out) {
     /* Create top-level graph object. */
     MVMSpeshGraph *g     = MVM_calloc(1, sizeof(MVMSpeshGraph));
     g->sf                = sf;
@@ -1259,6 +1325,7 @@ MVMSpeshGraph * MVM_spesh_graph_create_from_cand(MVMThreadContext *tc, MVMStatic
     g->num_deopt_addrs   = cand->num_deopts;
     g->alloc_deopt_addrs = cand->num_deopts;
     g->deopt_named_used_bit_field = cand->deopt_named_used_bit_field;
+    g->deopt_pea         = cand->deopt_pea;
     g->local_types       = cand->local_types;
     g->lexical_types     = cand->lexical_types;
     g->num_spesh_slots   = cand->num_spesh_slots;
@@ -1277,7 +1344,8 @@ MVMSpeshGraph * MVM_spesh_graph_create_from_cand(MVMThreadContext *tc, MVMStatic
     }
 
     /* Build the CFG out of the static frame, and transform it to SSA. */
-    build_cfg(tc, g, sf, cand->deopts, cand->num_deopts);
+    build_cfg(tc, g, sf, cand->deopts, cand->num_deopts, cand->deopt_usage_info,
+            deopt_usage_ins_out);
     if (!cfg_only) {
         MVM_spesh_eliminate_dead_bbs(tc, g, 0);
         add_predecessors(tc, g);
@@ -1307,7 +1375,7 @@ void MVM_spesh_graph_recompute_dominance(MVMThreadContext *tc, MVMSpeshGraph *g)
 
     /* Now form the new dominance tree. */
     add_predecessors(tc, g);
-    rpo = reverse_postorder(tc, g);
+    rpo = MVM_spesh_graph_reverse_postorder(tc, g);
     doms = compute_dominators(tc, g, rpo);
     add_children(tc, g, rpo, doms);
     MVM_free(rpo);
@@ -1350,12 +1418,64 @@ void MVM_spesh_graph_mark(MVMThreadContext *tc, MVMSpeshGraph *g, MVMGCWorklist
         MVM_gc_worklist_add(tc, worklist, &(g->inlines[i].sf));
 }
 
+/* Describes GCables for a heap snapshot. */
+void MVM_spesh_graph_describe(MVMThreadContext *tc, MVMSpeshGraph *g, MVMHeapSnapshotState *snapshot) {
+    MVMuint16 i, j, num_locals, num_facts, *local_types;
+
+    /* Mark static frame. */
+    MVM_profile_heap_add_collectable_rel_const_cstr(tc, snapshot, (MVMCollectable *)g->sf, "Static frame");
+
+    /* Mark facts. */
+    num_locals = g->num_locals;
+    local_types = g->local_types ? g->local_types : g->sf->body.local_types;
+    for (i = 0; i < num_locals; i++) {
+        num_facts = g->fact_counts[i];
+        for (j = 0; j < num_facts; j++) {
+            MVMint32 flags = g->facts[i][j].flags;
+            if (flags & MVM_SPESH_FACT_KNOWN_TYPE)
+                MVM_profile_heap_add_collectable_rel_const_cstr(tc, snapshot, (MVMCollectable *)g->facts[i][j].type, "Known Type");
+            if (flags & MVM_SPESH_FACT_KNOWN_DECONT_TYPE)
+                MVM_profile_heap_add_collectable_rel_const_cstr(tc, snapshot, (MVMCollectable *)g->facts[i][j].decont_type, "Known Decont Type");
+            if (flags & MVM_SPESH_FACT_KNOWN_VALUE) {
+                if (local_types[i] == MVM_reg_obj)
+                    MVM_profile_heap_add_collectable_rel_const_cstr(tc, snapshot, (MVMCollectable *)g->facts[i][j].value.o, "Known Value");
+                else if (local_types[i] == MVM_reg_str)
+                    MVM_profile_heap_add_collectable_rel_const_cstr(tc, snapshot, (MVMCollectable *)g->facts[i][j].value.s, "Known String Value");
+            }
+        }
+    }
+
+    /* Mark spesh slots. */
+    for (i = 0; i < g->num_spesh_slots; i++)
+        MVM_profile_heap_add_collectable_rel_idx(tc, snapshot, g->spesh_slots[i], i);
+
+    /* Mark inlines. */
+    for (i = 0; i < g->num_inlines; i++)
+        MVM_profile_heap_add_collectable_rel_idx(tc, snapshot, (MVMCollectable *)g->inlines[i].sf, i);
+}
+
 /* Destroys a spesh graph, deallocating all its associated memory. */
 void MVM_spesh_graph_destroy(MVMThreadContext *tc, MVMSpeshGraph *g) {
     /* Free all of the allocated node memory. */
     MVM_region_destroy(tc, &g->region_alloc);
-    /* Free handlers array, if different from the static frame. */
-    if (!g->cand && g->handlers && g->handlers != g->sf->body.handlers)
+    /* If there is a candidate that we either generated or that this graph was
+     * generated from, it has ownership of the malloc'd memory. If not, then we
+     * need to clean up */
+    if (g->spesh_slots && (!g->cand || g->cand->spesh_slots != g->spesh_slots))
+        MVM_free(g->spesh_slots);
+    if (g->deopt_addrs && (!g->cand || g->cand->deopts != g->deopt_addrs))
+        MVM_free(g->deopt_addrs);
+    if (g->inlines && (!g->cand || g->cand->inlines != g->inlines))
+        MVM_free(g->inlines);
+    if (g->local_types &&  (!g->cand || g->cand->local_types != g->local_types))
+        MVM_free(g->local_types);
+    if (g->lexical_types &&  (!g->cand || g->cand->lexical_types != g->lexical_types))
+        MVM_free(g->lexical_types);
+
+    /* Handlers can come directly from static frame, from spesh candidate, and
+     * from malloc/realloc. We only free it in the last case */
+    if (g->handlers && g->handlers != g->sf->body.handlers &&
+        (!g->cand || g->cand->handlers != g->handlers))
         MVM_free(g->handlers);
 
     /* Free the graph itself. */
