@@ -69,6 +69,9 @@ struct MVMSpeshGraph {
     MVMuint32 *always_retained_deopt_idxs;
     MVMuint32 num_always_retained_deopt_idxs;
 
+    /* Deopt information produced by escape analysis and scalar replacement. */
+    MVMSpeshPEADeopt deopt_pea;
+
     /* Table of information about inlines, laid out in order of nesting
      * depth. Thus, going through the table in order and finding when we
      * are within the bounds will show up each call frame that needs to
@@ -110,6 +113,9 @@ struct MVMSpeshGraph {
 
     /* Did we specialize on the invocant type? */
     MVMuint8 specialized_on_invocant;
+
+    /* Do we set a dispatcher? */
+    MVMuint8 sets_dispatcher;
 };
 
 /* A temporary register, added to support transformations. */
@@ -245,6 +251,7 @@ struct MVMSpeshAnn {
             MVMuint32 filename_string_index;
             MVMuint32 line_number;
         } lineno;
+        char *comment;
     } data;
 };
 
@@ -261,22 +268,28 @@ struct MVMSpeshAnn {
 #define MVM_SPESH_ANN_LINENO        10
 #define MVM_SPESH_ANN_LOGGED        11
 #define MVM_SPESH_ANN_DEOPT_SYNTH   12
+#define MVM_SPESH_ANN_COMMENT       4096
 
 /* Functions to create/destroy the spesh graph. */
 MVMSpeshGraph * MVM_spesh_graph_create(MVMThreadContext *tc, MVMStaticFrame *sf,
     MVMuint32 cfg_only, MVMuint32 insert_object_nulls);
 MVMSpeshGraph * MVM_spesh_graph_create_from_cand(MVMThreadContext *tc, MVMStaticFrame *sf,
-    MVMSpeshCandidate *cand, MVMuint32 cfg_only);
+    MVMSpeshCandidate *cand, MVMuint32 cfg_only, MVMSpeshIns ***deopt_usage_ins_out);
 MVMSpeshBB * MVM_spesh_graph_linear_prev(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *search);
 void MVM_spesh_graph_grow_deopt_table(MVMThreadContext *tc, MVMSpeshGraph *g);
 void MVM_spesh_graph_add_deopt_annotation(MVMThreadContext *tc, MVMSpeshGraph *g,
     MVMSpeshIns *ins_node, MVMuint32 deopt_target, MVMint32 type);
+MVMSpeshBB ** MVM_spesh_graph_reverse_postorder(MVMThreadContext *tc, MVMSpeshGraph *g);
 void MVM_spesh_graph_recompute_dominance(MVMThreadContext *tc, MVMSpeshGraph *g);
 void MVM_spesh_graph_mark(MVMThreadContext *tc, MVMSpeshGraph *g, MVMGCWorklist *worklist);
+void MVM_spesh_graph_describe(MVMThreadContext *tc, MVMSpeshGraph *g, MVMHeapSnapshotState *snapshot);
 void MVM_spesh_graph_destroy(MVMThreadContext *tc, MVMSpeshGraph *g);
 MVM_PUBLIC void * MVM_spesh_alloc(MVMThreadContext *tc, MVMSpeshGraph *g, size_t bytes);
 MVMOpInfo *get_phi(MVMThreadContext *tc, MVMSpeshGraph *g, MVMuint32 nrargs);
 
+MVM_PUBLIC void MVM_spesh_graph_add_comment(MVMThreadContext *tc, MVMSpeshGraph *g,
+    MVMSpeshIns *ins, const char *fmt, ...);
+
 MVM_STATIC_INLINE MVMuint32 MVM_spesh_is_inc_dec_op(MVMuint16 opcode) {
     return opcode == MVM_OP_inc_i || opcode == MVM_OP_dec_i ||
            opcode == MVM_OP_inc_u || opcode == MVM_OP_dec_u;
