@@ -83,6 +83,16 @@ static int is_static_frame_inlineable(MVMThreadContext *tc, MVMSpeshGraph *inlin
         return 0;
     }
 
+    /* Ensure that we haven't hit size limits. */
+    if (inliner->num_locals > MVM_SPESH_INLINE_MAX_LOCALS) {
+        *no_inline_reason = "inliner has too many locals";
+        return 0;
+    }
+    if (inliner->num_inlines > MVM_SPESH_INLINE_MAX_INLINES) {
+        *no_inline_reason = "inliner has too many inlines";
+        return 0;
+    }
+
     return 1;
 }
 
@@ -91,7 +101,7 @@ static int is_static_frame_inlineable(MVMThreadContext *tc, MVMSpeshGraph *inlin
  * and returns zero. Optionally builds up usage counts on the graph. */
 static int is_graph_inlineable(MVMThreadContext *tc, MVMSpeshGraph *inliner,
                                MVMStaticFrame *target_sf, MVMSpeshIns *invoke_ins,
-                               MVMSpeshGraph *ig, char **no_inline_reason, MVMOpInfo **no_inline_info) {
+                               MVMSpeshGraph *ig, char **no_inline_reason, MVMOpInfo const **no_inline_info) {
     MVMSpeshBB *bb = ig->entry;
     MVMint32 same_hll = target_sf->body.cu->body.hll_config ==
             inliner->sf->body.cu->body.hll_config;
@@ -211,6 +221,12 @@ void add_deopt_usages(MVMThreadContext *tc, MVMSpeshGraph *g, MVMint32 *deopt_us
     }
 }
 
+/* Get the maximum inline size applicable to the specified static frame (it can
+ * be configured by language). */
+int MVM_spesh_inline_get_max_size(MVMThreadContext *tc, MVMStaticFrame *sf) {
+    return sf->body.cu->body.hll_config->max_inline_size;
+}
+
 /* Sees if it will be possible to inline the target code ref, given we could
  * already identify a spesh candidate. Returns NULL if no inlining is possible
  * or a graph ready to be merged if it will be possible. */
@@ -220,13 +236,13 @@ MVMSpeshGraph * MVM_spesh_inline_try_get_graph(MVMThreadContext *tc, MVMSpeshGra
                                                MVMSpeshIns *invoke_ins,
                                                char **no_inline_reason,
                                                MVMuint32 *effective_size,
-                                               MVMOpInfo **no_inline_info) {
+                                               MVMOpInfo const **no_inline_info) {
     MVMSpeshGraph *ig;
     MVMSpeshIns **deopt_usage_ins = NULL;
 
     /* Check bytecode size is within the inline limit. */
     *effective_size = get_effective_size(tc, cand);
-    if (*effective_size > MVM_SPESH_MAX_INLINE_SIZE) {
+    if (*effective_size > MVM_spesh_inline_get_max_size(tc, target_sf)) {
         *no_inline_reason = "bytecode is too large to inline";
         return NULL;
     }
@@ -268,7 +284,7 @@ MVMSpeshGraph * MVM_spesh_inline_try_get_graph(MVMThreadContext *tc, MVMSpeshGra
 /* Tries to get a spesh graph for a particular unspecialized candidate. */
 MVMSpeshGraph * MVM_spesh_inline_try_get_graph_from_unspecialized(MVMThreadContext *tc,
         MVMSpeshGraph *inliner, MVMStaticFrame *target_sf, MVMSpeshIns *invoke_ins,
-        MVMSpeshCallInfo *call_info, MVMSpeshStatsType *type_tuple, char **no_inline_reason, MVMOpInfo **no_inline_info) {
+        MVMSpeshCallInfo *call_info, MVMSpeshStatsType *type_tuple, char **no_inline_reason, MVMOpInfo const **no_inline_info) {
     MVMSpeshGraph *ig;
 
     /* Cannot inline with flattening args. */
@@ -435,6 +451,22 @@ static void rewrite_outer_lookup(MVMThreadContext *tc, MVMSpeshGraph *g,
     MVM_spesh_usages_add_by_reg(tc, g, code_ref_reg, ins);
 }
 
+/* Rewrites a lexical bind to an outer to be done via. a register holding
+ * the outer coderef. */
+static void rewrite_outer_bind(MVMThreadContext *tc, MVMSpeshGraph *g,
+                               MVMSpeshIns *ins, MVMuint16 num_locals,
+                               MVMuint16 op, MVMSpeshOperand code_ref_reg) {
+    MVMSpeshOperand *new_operands = MVM_spesh_alloc(tc, g, 4 * sizeof(MVMSpeshOperand));
+    new_operands[0].lit_ui16 = ins->operands[0].lex.idx;
+    new_operands[1].lit_ui16 = ins->operands[0].lex.outers;
+    new_operands[2] = code_ref_reg;
+    new_operands[3] = ins->operands[1];
+    new_operands[3].reg.orig += num_locals;
+    ins->info = MVM_op_get_op(op);
+    ins->operands = new_operands;
+    MVM_spesh_usages_add_by_reg(tc, g, code_ref_reg, ins);
+}
+
 static void rewrite_hlltype(MVMThreadContext *tc, MVMSpeshGraph *inlinee, MVMSpeshIns *ins) {
     MVMObject *selected_type;
     MVMHLLConfig *hll = inlinee->sf->body.cu->body.hll_config;
@@ -546,6 +578,14 @@ MVMSpeshBB * merge_graph(MVMThreadContext *tc, MVMSpeshGraph *inliner,
                 rewrite_outer_lookup(tc, inliner, ins, inliner->num_locals,
                     MVM_OP_sp_getlexvia_ins, code_ref_reg);
             }
+            else if (opcode == MVM_OP_sp_bindlex_in && ins->operands[0].lex.outers > 0) {
+                rewrite_outer_bind(tc, inliner, ins, inliner->num_locals,
+                    MVM_OP_sp_bindlexvia_in, code_ref_reg);
+            }
+            else if (opcode == MVM_OP_sp_bindlex_os && ins->operands[0].lex.outers > 0) {
+                rewrite_outer_bind(tc, inliner, ins, inliner->num_locals,
+                    MVM_OP_sp_bindlexvia_os, code_ref_reg);
+            }
             else if (opcode == MVM_OP_getlex && ins->operands[1].lex.outers > 0) {
                 MVMuint16 outers = ins->operands[1].lex.outers;
                 MVMStaticFrame *outer = inlinee_sf;
@@ -623,17 +663,6 @@ MVMSpeshBB * merge_graph(MVMThreadContext *tc, MVMSpeshGraph *inliner,
     }
     MVM_VECTOR_DESTROY(regs_for_deopt);
 
-    /* If we saw that we may deopt anywhere, mark the basic blocks of the
-     * inline as maybe deopting; this prevents us doing too aggressive
-     * optimizations within them. */
-    if (may_cause_deopt) {
-        bb = inlinee->entry;
-        while (bb) {
-            bb->inlined_may_cause_deopt = 1;
-            bb = bb->linear_next;
-        }
-    }
-
     /* Link inlinee BBs into the linear next chain. */
     bb = invoke_bb->linear_next;
     invoke_bb->linear_next = inlinee_first_bb = inlinee->entry->linear_next;
