@@ -8,7 +8,9 @@
 #define MVM_encoding_type_utf8_c8       6
 #define MVM_encoding_type_windows1251   7
 #define MVM_encoding_type_shiftjis      8
-#define MVM_encoding_type_MAX           8
+#define MVM_encoding_type_utf16le       9
+#define MVM_encoding_type_utf16be      10
+#define MVM_encoding_type_MAX          10
 #define ENCODING_VALID(enc) \
     (((enc) >= MVM_encoding_type_MIN && (enc) <= MVM_encoding_type_MAX) \
     || (MVM_exception_throw_adhoc(tc, "invalid encoding type flag: %d", (enc)),1))
@@ -73,6 +75,8 @@ MVM_STATIC_INLINE int MVM_string_buf32_can_fit_into_8bit(MVMGrapheme32 *active_b
 }
 MVMGrapheme32 MVM_string_get_grapheme_at_nocheck(MVMThreadContext *tc, MVMString *a, MVMint64 index);
 MVMint64 MVM_string_equal(MVMThreadContext *tc, MVMString *a, MVMString *b);
+MVMint64 MVM_string_substrings_equal_nocheck(MVMThreadContext *tc, MVMString *a,
+        MVMint64 starta, MVMint64 length, MVMString *b, MVMint64 startb);
 MVMint64 MVM_string_index(MVMThreadContext *tc, MVMString *haystack, MVMString *needle, MVMint64 start);
 MVMint64 MVM_string_index_ignore_case(MVMThreadContext *tc, MVMString *haystack, MVMString *needle, MVMint64 start);
 MVMint64 MVM_string_index_ignore_mark(MVMThreadContext *tc, MVMString *Haystack, MVMString *needle, MVMint64 start);
@@ -129,6 +133,7 @@ MVMString * MVM_string_chr(MVMThreadContext *tc, MVMint64 cp);
 MVMint64 MVM_string_grapheme_is_cclass(MVMThreadContext *tc, MVMint64 cclass, MVMGrapheme32 g);
 void MVM_string_compute_hash_code(MVMThreadContext *tc, MVMString *s);
 MVMString * MVM_string_ascii_from_buf_nocheck(MVMThreadContext *tc, MVMGrapheme8 *buf, MVMStringIndex len);
+char * MVM_string_encoding_cname(MVMThreadContext *tc, MVMint64 encoding);
 /* If MVM_DEBUG_NFG is 1, calls to NFG_CHECK will re_nfg the given string
  * and compare num_graphs before and after the normalization.
  * If it is different debug information will be printed out.*/
