@@ -206,6 +206,7 @@ typedef struct MVMSpeshPEAAllocation MVMSpeshPEAAllocation;
 typedef struct MVMSpeshPEADeopt MVMSpeshPEADeopt;
 typedef struct MVMSpeshPEAMaterializeInfo MVMSpeshPEAMaterializeInfo;
 typedef struct MVMSpeshPEADeoptPoint MVMSpeshPEADeoptPoint;
+typedef struct MVMConfigurationProgram MVMConfigurationProgram;
 typedef struct MVMSTable MVMSTable;
 typedef struct MVMStaticFrame MVMStaticFrame;
 typedef struct MVMStaticFrameBody MVMStaticFrameBody;
@@ -274,6 +275,7 @@ typedef struct MVMProfileThreadData MVMProfileThreadData;
 typedef struct MVMProfileGC MVMProfileGC;
 typedef struct MVMProfileCallNode MVMProfileCallNode;
 typedef struct MVMProfileAllocationCount MVMProfileAllocationCount;
+typedef struct MVMProfileDeallocationCount MVMProfileDeallocationCount;
 typedef struct MVMProfileContinuationData MVMProfileContinuationData;
 typedef struct MVMHeapSnapshotCollection MVMHeapSnapshotCollection;
 typedef struct MVMHeapDumpIndex MVMHeapDumpIndex;
